from functools import cache

import pytest
from rdflib import RDF
from rdflib import RDFS
from rdflib import SH
from rdflib import Graph
from rdflib import Literal

from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import ValidationResult
from dsp_tools.commands.validate_data.models.validation import ValidationResultBaseInfo
from dsp_tools.commands.validate_data.models.validation import ViolationType
from dsp_tools.utils.rdflib_constants import DASH
from dsp_tools.utils.rdflib_constants import DATA
from dsp_tools.utils.rdflib_constants import KNORA_API
from test.unittests.commands.validate_data.constants import IN_BUILT_ONTO
from test.unittests.commands.validate_data.constants import ONTO
from test.unittests.commands.validate_data.constants import PREFIXES


@cache
@pytest.fixture
def onto_graph() -> Graph:
    g = Graph()
    g.parse("testdata/validate-data/onto.ttl")
    g.parse("testdata/validate-data/knora-api-subset.ttl")
    return g


@pytest.fixture
def report_target_resource_wrong_type(onto_graph: Graph) -> tuple[Graph, Graph]:
    validation_str = f"""{PREFIXES}
    [ 
        a sh:ValidationResult ;
        sh:detail _:detail_bn ;
        sh:focusNode <http://data/region_isRegionOf_resource_not_a_representation> ;
        sh:resultMessage "Value does not have shape api-shapes:isRegionOf_NodeShape" ;
        sh:resultPath <http://api.knora.org/ontology/knora-api/v2#isRegionOf> ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:NodeConstraintComponent ;
        sh:sourceShape <http://api.knora.org/ontology/knora-api/shapes/v2#isRegionOf_PropertyShape> ;
        sh:value <http://data/value_isRegionOf> 
    ] .
    
    _:detail_bn a sh:ValidationResult ;
        sh:focusNode <http://data/value_isRegionOf> ;
        sh:resultMessage "http://api.knora.org/ontology/knora-api/v2#Representation" ;
        sh:resultPath <http://api.knora.org/ontology/knora-api/shapes/v2#linkValueHasTargetID> ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:ClassConstraintComponent ;
        sh:sourceShape _:source_shape ;
        sh:value <http://data/target_res_without_representation_1> .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/region_isRegionOf_resource_not_a_representation> 
        a knora-api:Region ;
        rdfs:label "Region"^^xsd:string ;
        knora-api:hasColor <http://data/value_hasColor> ;
        knora-api:hasGeometry <http://data/value_hasGeometry> ;
        knora-api:isRegionOf <http://data/value_isRegionOf> .
    
    <http://data/value_isRegionOf> a knora-api:LinkValue ;
        api-shapes:linkValueHasTargetID <http://data/target_res_without_representation_1> .
    
    <http://data/target_res_without_representation_1> a in-built:TestNormalResource ;
        rdfs:label "Resource without Representation"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    return validation_g, onto_data_g


@pytest.fixture
def report_not_resource(onto_graph: Graph) -> tuple[Graph, Graph]:
    validation_str = f"""{PREFIXES}
    _:bn_id_simpletext a sh:ValidationResult ;
        sh:focusNode <http://data/value_id_simpletext> ;
        sh:resultMessage "TextValue without formatting" ;
        sh:resultPath knora-api:valueAsString ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
        sh:sourceShape api-shapes:SimpleTextValue_PropShape .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/id_simpletext> a onto:ClassWithEverything ;
        rdfs:label "Simpletext"^^xsd:string ;
        onto:testTextarea <http://data/value_id_simpletext> .
    <http://data/value_id_simpletext> a knora-api:TextValue ;
        knora-api:textValueAsXml "Text"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    return validation_g, onto_data_g


@pytest.fixture
def report_min_card(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/id_card_one> ;
        sh:resultMessage "1" ;
        sh:resultPath onto:testBoolean ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
        sh:sourceShape [ ] ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/id_card_one> a onto:ClassInheritedCardinalityOverwriting ;
        rdfs:label "Bool Card 1"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.MinCountConstraintComponent,
        focus_node_iri=DATA.id_card_one,
        focus_node_type=ONTO.ClassInheritedCardinalityOverwriting,
        result_path=ONTO.testBoolean,
        severity=SH.Violation,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def file_value_cardinality_to_ignore(onto_graph: Graph) -> tuple[Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
[   a sh:ValidationResult ;
    sh:focusNode <http://data/id_wrong_file_type> ;
    sh:resultMessage "Property knora-api:hasMovingImageFileValue is not among those permitted for any of the types" ;
    sh:resultPath <http://api.knora.org/ontology/knora-api/v2#hasMovingImageFileValue> ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent <http://datashapes.org/dash#ClosedByTypesConstraintComponent> ;
    sh:sourceShape <http://0.0.0.0:3333/ontology/9999/onto/v2#TestStillImageRepresentation> ;
    sh:value <http://data/fileValueBn> ] .
    """
    data_str = f"""{PREFIXES}
    <http://data/id_wrong_file_type> a onto:TestStillImageRepresentation ;
        rdfs:label "TestStillImageRepresentation File mp4"^^xsd:string ;
        knora-api:hasMovingImageFileValue <http://data/fileValueBn> .
    """
    graphs = Graph()
    graphs.parse(data=validation_str, format="ttl")
    graphs.parse(data=data_str, format="ttl")
    graphs += onto_graph
    val_bn = next(graphs.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=DASH.ClosedByTypesConstraintComponent,
        focus_node_iri=DATA.id_wrong_file_type,
        focus_node_type=ONTO.TestStillImageRepresentation,
        result_path=KNORA_API.hasMovingImageFileValue,
        severity=SH.Violation,
    )
    return graphs, base_info


@pytest.fixture
def file_value_for_resource_without_representation(onto_graph: Graph) -> tuple[Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
[ a sh:ValidationResult ;
    sh:focusNode <http://data/id_resource_without_representation> ;
    sh:resultMessage "Property knora-api:hasMovingImageFileValue is not among those permitted for any of the types" ;
    sh:resultPath <http://api.knora.org/ontology/knora-api/v2#hasMovingImageFileValue> ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent <http://datashapes.org/dash#ClosedByTypesConstraintComponent> ;
    sh:sourceShape <http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything> ;
    sh:value <http://data/fileBn> ] .
    """
    data_str = f"""{PREFIXES}
    <http://data/id_resource_without_representation> a <http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything> ;
        rdfs:label "Resource Without Representation"^^xsd:string ;
        knora-api:hasMovingImageFileValue <http://data/fileBn> .
    """
    graphs = Graph()
    graphs.parse(data=validation_str, format="ttl")
    graphs.parse(data=data_str, format="ttl")
    graphs += onto_graph
    val_bn = next(graphs.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=DASH.ClosedByTypesConstraintComponent,
        focus_node_iri=DATA.id_resource_without_representation,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=KNORA_API.hasMovingImageFileValue,
        severity=SH.Violation,
    )
    return graphs, base_info


@pytest.fixture
def extracted_file_value_for_resource_without_representation() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.FILEVALUE_PROHIBITED,
        res_iri=DATA.id_resource_without_representation,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.hasMovingImageFileValue,
        severity=SH.Violation,
    )


@pytest.fixture
def report_file_value_duplicate(onto_graph: Graph) -> tuple[Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
[ a sh:ValidationResult ;
    sh:focusNode <http://data/value_duplicate_archive_1> ;
    sh:resultMessage "The entered filepath is used more than once in your data." ;
    sh:resultPath <http://api.knora.org/ontology/knora-api/v2#fileValueHasFilename> ;
    sh:resultSeverity sh:Info ;
    sh:sourceConstraintComponent <http://datashapes.org/dash#UniqueValueForClassConstraintComponent> ;
    sh:sourceShape _:nbc536211f72c45abab34094e81552f87b26 ;
    sh:value "duplicate_file.zip" ] .
    """
    data_str = f"""{PREFIXES}
<http://data/duplicate_archive_1> a <http://0.0.0.0:3333/ontology/9999/onto/v2#TestArchiveRepresentation> ;
    rdfs:label "duplicate file"^^xsd:string ;
    knora-api:hasArchiveFileValue <http://data/value_duplicate_archive_1> .
    
<http://data/value_duplicate_archive_1> a knora-api:ArchiveFileValue ;
    knora-api:fileValueHasFilename "duplicate_file.zip"^^xsd:string .
    """
    graphs = Graph()
    graphs.parse(data=validation_str, format="ttl")
    graphs.parse(data=data_str, format="ttl")
    graphs += onto_graph
    val_bn = next(graphs.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=DASH.UniqueValueForClassConstraintComponent,
        focus_node_iri=DATA.duplicate_archive_1,
        focus_node_type=ONTO.TestArchiveRepresentation,
        result_path=KNORA_API.hasArchiveFileValue,
        severity=SH.Info,
    )
    return graphs, base_info


@pytest.fixture
def extracted_file_value_duplicate() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.FILE_DUPLICATE,
        res_iri=DATA.duplicate_archive_1,
        res_class=ONTO.TestArchiveRepresentation,
        property=KNORA_API.hasArchiveFileValue,
        input_value=Literal("duplicate_file.zip"),
        message=Literal("The entered filepath is used more than once in your data."),
        severity=SH.Info,
    )


@pytest.fixture
def extracted_min_card() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.MIN_CARD,
        res_iri=DATA.id_card_one,
        res_class=ONTO.ClassInheritedCardinalityOverwriting,
        property=ONTO.testBoolean,
        expected=Literal("1"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_value_type_simpletext(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:detail _:bn_id_simpletext ;
        sh:focusNode <http://data/id_simpletext> ;
        sh:resultMessage "Value does not have shape <http://api.knora.org/ontology/knora-api/shapes/v2#SimpleTextValue_ClassShape>" ;
        sh:resultPath onto:testTextarea ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:NodeConstraintComponent ;
        sh:sourceShape onto:testTextarea_PropShape ;
        sh:value <http://data/value_id_simpletext> ] .

    _:bn_id_simpletext a sh:ValidationResult ;
        sh:focusNode <http://data/value_id_simpletext> ;
        sh:resultMessage "TextValue without formatting" ;
        sh:resultPath knora-api:valueAsString ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
        sh:sourceShape api-shapes:SimpleTextValue_PropShape .
    """  # noqa: E501 (Line too long)
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/id_simpletext> a onto:ClassWithEverything ;
        rdfs:label "Simpletext"^^xsd:string ;
        onto:testTextarea <http://data/value_id_simpletext> .

    <http://data/value_id_simpletext> a knora-api:TextValue ;
        knora-api:textValueAsXml "Text"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    detail_bn = next(validation_g.objects(val_bn, SH.detail))
    detail = DetailBaseInfo(
        detail_bn=detail_bn,
        source_constraint_component=SH.MinCountConstraintComponent,
    )
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        result_path=ONTO.testTextarea,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.id_simpletext,
        focus_node_type=ONTO.ClassWithEverything,
        severity=SH.Violation,
        detail=detail,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_value_type_simpletext() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.VALUE_TYPE,
        res_iri=DATA.id_simpletext,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testTextarea,
        expected=Literal("TextValue without formatting"),
        input_type=KNORA_API.TextValue,
        severity=SH.Violation,
    )


@pytest.fixture
def report_min_inclusive(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/value_iri> ;
        sh:resultMessage "The interval start must be a non-negative integer or decimal." ;
        sh:resultPath <http://api.knora.org/ontology/knora-api/v2#intervalValueHasStart> ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:MinInclusiveConstraintComponent ;
        sh:sourceShape <http://api.knora.org/ontology/knora-api/shapes/v2#intervalValueHasStart_PropShape> ;
        sh:value -2.0 ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}

    <http://data/video_segment_negative_bounds> a knora-api:VideoSegment ;
        rdfs:label "Video Segment"^^xsd:string ;
        knora-api:hasSegmentBounds <http://data/value_iri> .
    
    <http://data/value_iri> a knora-api:IntervalValue ;
        knora-api:intervalValueHasEnd -1.0 ;
        knora-api:intervalValueHasStart -2.0 .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        result_path=KNORA_API.hasSegmentBounds,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.video_segment_negative_bounds,
        focus_node_type=KNORA_API.VideoSegment,
        severity=SH.Violation,
        detail=None,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_min_inclusive() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.GENERIC,
        res_iri=DATA.video_segment_negative_bounds,
        res_class=KNORA_API.VideoSegment,
        property=KNORA_API.hasSegmentBounds,
        message=Literal("The interval start must be a non-negative integer or decimal."),
        input_value=Literal("-2.0"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_value_type(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/id_uri> ;
        sh:resultMessage "This property requires a UriValue" ;
        sh:resultPath <http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue> ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:ClassConstraintComponent ;
        sh:sourceShape [ ] ;
        sh:value <http://data/value_id_uri> ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/id_uri> a onto:ClassWithEverything ;
        rdfs:label "Uri"^^xsd:string ;
        onto:testUriValue <http://data/value_id_uri> .

    <http://data/value_id_uri> a knora-api:TextValue ;
        knora-api:valueAsString "https://dasch.swiss"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.ClassConstraintComponent,
        focus_node_iri=DATA.id_uri,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testUriValue,
        severity=SH.Violation,
        detail=None,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_value_type() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.VALUE_TYPE,
        res_iri=DATA.id_uri,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testUriValue,
        expected=Literal("This property requires a UriValue"),
        input_type=KNORA_API.TextValue,
        severity=SH.Violation,
    )


@pytest.fixture
def report_regex(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/value_geoname_not_number> ;
        sh:resultMessage "The value must be a valid geoname code" ;
        sh:resultPath <http://api.knora.org/ontology/knora-api/v2#geonameValueAsGeonameCode> ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:PatternConstraintComponent ;
        sh:sourceShape <http://api.knora.org/ontology/knora-api/shapes/v2#geonameValueAsGeonameCode_Shape> ;
        sh:value "this-is-not-a-valid-code" ].
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/geoname_not_number> a onto:ClassWithEverything ;
        rdfs:label "Geoname is not a number"^^xsd:string ;
        onto:testGeoname <http://data/value_geoname_not_number> .

    <http://data/value_geoname_not_number> a knora-api:GeonameValue ;
        knora-api:geonameValueAsGeonameCode "this-is-not-a-valid-code"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.geoname_not_number,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testGeoname,
        severity=SH.Violation,
        detail=None,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_regex() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.PATTERN,
        res_iri=DATA.geoname_not_number,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testGeoname,
        expected=Literal("The value must be a valid geoname code"),
        input_value=Literal("this-is-not-a-valid-code"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_link_target_non_existent(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:detail _:bn_link_target_non_existent ;
        sh:focusNode <http://data/link_target_non_existent> ;
        sh:resultMessage "Value does not have shape <http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo_NodeShape>" ;
        sh:resultPath onto:testHasLinkTo ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:NodeConstraintComponent ;
        sh:sourceShape onto:testHasLinkTo_PropShape ;
        sh:value <http://data/value_link_target_non_existent> ] .
    
    _:bn_link_target_non_existent a sh:ValidationResult ;
        sh:focusNode <http://data/value_link_target_non_existent> ;
        sh:resultMessage <http://api.knora.org/ontology/knora-api/v2#Resource> ;
        sh:resultPath api-shapes:linkValueHasTargetID ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:ClassConstraintComponent ;
        sh:sourceShape [ ] ;
        sh:value <http://data/other> .
    """  # noqa: E501 (Line too long)
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/link_target_non_existent> a onto:ClassWithEverything ;
        rdfs:label "Target does not exist"^^xsd:string ;
        onto:testHasLinkTo <http://data/value_link_target_non_existent> .
    
    <http://data/value_link_target_non_existent> a knora-api:LinkValue ;
    api-shapes:linkValueHasTargetID <http://data/other> .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    detail_bn = next(validation_g.objects(val_bn, SH.detail))
    detail_component = next(validation_g.objects(detail_bn, SH.sourceConstraintComponent))
    detail = DetailBaseInfo(
        detail_bn=detail_bn,
        source_constraint_component=detail_component,
    )
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.link_target_non_existent,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testHasLinkTo,
        severity=SH.Violation,
        detail=detail,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_link_target_non_existent() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.LINK_TARGET,
        res_iri=DATA.link_target_non_existent,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testHasLinkTo,
        expected=KNORA_API.Resource,
        input_value=DATA.other,
        severity=SH.Violation,
    )


@pytest.fixture
def report_link_target_wrong_class(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:detail _:bn_link_target_wrong_class ;
        sh:focusNode <http://data/link_target_wrong_class> ;
        sh:resultMessage "Value does not have shape <http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkToCardOneResource_NodeShape>" ;
        sh:resultPath onto:testHasLinkToCardOneResource ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:NodeConstraintComponent ;
        sh:sourceShape onto:testHasLinkToCardOneResource_PropShape ;
        sh:value <http://data/value_link_target_wrong_class> ] .
        
    _:bn_link_target_wrong_class a sh:ValidationResult ;
        sh:focusNode <http://data/value_link_target_wrong_class> ;
        sh:resultMessage <http://0.0.0.0:3333/ontology/9999/onto/v2#CardOneResource> ;
        sh:resultPath api-shapes:linkValueHasTargetID ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:ClassConstraintComponent ;
        sh:sourceShape [ ] ;
        sh:value <http://data/id_9_target> .
    """  # noqa: E501 (Line too long)
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/link_target_wrong_class> a onto:ClassWithEverything ;
        rdfs:label "Target not the right class"^^xsd:string ;
        onto:testHasLinkToCardOneResource <http://data/value_link_target_wrong_class> .
    
    <http://data/id_9_target> a onto:ClassWithEverything ;
        rdfs:label "Link Prop"^^xsd:string .

    <http://data/value_link_target_wrong_class> a knora-api:LinkValue ;
        api-shapes:linkValueHasTargetID <http://data/id_9_target> .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    detail_bn = next(validation_g.objects(val_bn, SH.detail))
    detail_component = next(validation_g.objects(detail_bn, SH.sourceConstraintComponent))
    detail = DetailBaseInfo(
        detail_bn=detail_bn,
        source_constraint_component=detail_component,
    )
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.link_target_wrong_class,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testHasLinkToCardOneResource,
        severity=SH.Violation,
        detail=detail,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_link_target_wrong_class() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.LINK_TARGET,
        res_iri=DATA.link_target_wrong_class,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testHasLinkToCardOneResource,
        expected=ONTO.CardOneResource,
        input_value=DATA.id_9_target,
        input_type=ONTO.ClassWithEverything,
        severity=SH.Violation,
    )


@pytest.fixture
def report_image_missing_legal_info(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/value_image_no_legal_info> ;
        sh:resultMessage "Files and IIIF-URIs require a reference to a license." ;
        sh:resultPath <http://api.knora.org/ontology/knora-api/v2#hasLicense> ;
        sh:resultSeverity sh:Warning ;
        sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
        sh:sourceShape <http://api.knora.org/ontology/knora-api/shapes/v2#hasLicense_PropShape> ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/image_no_legal_info> a <http://0.0.0.0:3333/ontology/9999/onto/v2#TestStillImageRepresentation> ;
    rdfs:label "TestStillImageRepresentation Bitstream"^^xsd:string ;
    knora-api:hasStillImageFileValue <http://data/value_image_no_legal_info> .

    <http://data/value_image_no_legal_info> a knora-api:StillImageFileValue ;
        knora-api:fileValueHasFilename "this/is/filepath/file.jp2"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.MinCountConstraintComponent,
        focus_node_iri=DATA.image_no_legal_info,
        focus_node_type=ONTO.TestStillImageRepresentation,
        result_path=KNORA_API.hasLicense,
        severity=SH.Warning,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_image_missing_legal_info() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.GENERIC,
        res_iri=DATA.image_no_legal_info,
        res_class=ONTO.TestStillImageRepresentation,
        severity=SH.Warning,
        property=KNORA_API.hasLicense,
        expected=Literal("Files and IIIF-URIs require a reference to a license."),
    )


@pytest.fixture
def report_archive_missing_legal_info(onto_graph: Graph) -> tuple[Graph, Graph]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/value_bitstream_no_legal_info> ;
        sh:resultMessage "Files and IIIF-URIs require a reference to a license." ;
        sh:resultPath <http://api.knora.org/ontology/knora-api/v2#hasLicense> ;
        sh:resultSeverity sh:Warning ;
        sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
        sh:sourceShape <http://api.knora.org/ontology/knora-api/shapes/v2#hasLicense_PropShape> ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/bitstream_no_legal_info> a <http://0.0.0.0:3333/ontology/9999/onto/v2#TestArchiveRepresentation> ;
        rdfs:label "TestArchiveRepresentation tar"^^xsd:string ;
        knora-api:hasArchiveFileValue <http://data/value_bitstream_no_legal_info> .
    
    <http://data/value_bitstream_no_legal_info> a knora-api:ArchiveFileValue ;
        knora-api:fileValueHasFilename "this/is/filepath/file.tar"^^xsd:string .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    return validation_g, onto_data_g


@pytest.fixture
def report_closed_constraint(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/id_closed_constraint> ;
        sh:resultMessage "Property onto:testIntegerSimpleText is not among those permitted for any of the types" ;
        sh:resultPath onto:testIntegerSimpleText ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent dash:ClosedByTypesConstraintComponent ;
        sh:sourceShape onto:CardOneResource ;
        sh:value <http://data/value_id_closed_constraint> ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/id_closed_constraint> a onto:CardOneResource ;
        rdfs:label "Int card does not exist"^^xsd:string ;
        onto:testIntegerSimpleText <http://data/value_id_closed_constraint> .
    <http://data/value_id_closed_constraint> a knora-api:IntValue ;
        knora-api:intValueAsInt 1 .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=DASH.ClosedByTypesConstraintComponent,
        focus_node_iri=DATA.id_closed_constraint,
        focus_node_type=ONTO.CardOneResource,
        result_path=ONTO.testIntegerSimpleText,
        severity=SH.Violation,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_closed_constraint() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.NON_EXISTING_CARD,
        res_iri=DATA.id_closed_constraint,
        res_class=ONTO.CardOneResource,
        property=ONTO.testIntegerSimpleText,
        severity=SH.Violation,
    )


@pytest.fixture
def report_max_card(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/id_max_card> ;
        sh:resultMessage "1" ;
        sh:resultPath onto:testHasLinkToCardOneResource ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:MaxCountConstraintComponent ;
        sh:sourceShape [ ] ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/id_max_card> a onto:ClassMixedCard ;
        rdfs:label "Decimal Card 0-1"^^xsd:string ;
        onto:testHasLinkToCardOneResource <http://data/value_1> , <http://data/value_2> .

    <http://data/value_1> a knora-api:LinkValue ;
        api-shapes:linkValueHasTargetID <http://data/id_card_one> .
    <http://data/value_2> a knora-api:LinkValue ;
        api-shapes:linkValueHasTargetID <http://data/id_closed_constraint> .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.MaxCountConstraintComponent,
        focus_node_iri=DATA.id_max_card,
        focus_node_type=ONTO.ClassMixedCard,
        result_path=ONTO.testHasLinkToCardOneResource,
        severity=SH.Violation,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_max_card() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.MAX_CARD,
        res_iri=DATA.id_max_card,
        res_class=ONTO.ClassMixedCard,
        property=ONTO.testDecimalSimpleText,
        expected=Literal("0-1"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_empty_label(onto_graph: Graph) -> tuple[Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/empty_label> ;
        sh:resultMessage "The label must be a non-empty string" ;
        sh:resultPath rdfs:label ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:PatternConstraintComponent ;
        sh:sourceShape [ ] ;
        sh:value " " ] .
    """
    data_str = f"""{PREFIXES}
        <http://data/empty_label> a onto:ClassWithEverything ;
            rdfs:label " "^^xsd:string .
    """
    graphs = Graph()
    graphs.parse(data=validation_str, format="ttl")
    graphs.parse(data=data_str, format="ttl")
    graphs += onto_graph
    val_bn = next(graphs.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.PatternConstraintComponent,
        focus_node_iri=DATA.empty_label,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=RDFS.label,
        severity=SH.Violation,
    )
    return graphs, base_info


@pytest.fixture
def extracted_empty_label() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.PATTERN,
        res_iri=DATA.empty_label,
        res_class=ONTO.ClassWithEverything,
        property=RDFS.label,
        expected=Literal("The label must be a non-empty string"),
        input_value=Literal(" "),
        severity=SH.Violation,
    )


@pytest.fixture
def report_unique_value_literal(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/identical_values_valueHas> ;
        sh:resultMessage "A resource may not have the same property and value more than one time." ;
        sh:resultPath onto:testGeoname ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraint _:1 ;
        sh:sourceConstraintComponent sh:SPARQLConstraintComponent ;
        sh:sourceShape onto:ClassWithEverything_Unique ;
        sh:value "00111111" ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
        <http://data/identical_values_valueHas> a onto:ClassWithEverything .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.SPARQLConstraintComponent,
        focus_node_iri=DATA.identical_values_valueHas,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testGeoname,
        severity=SH.Violation,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_unique_value_literal() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.UNIQUE_VALUE,
        res_iri=DATA.identical_values_valueHas,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testGeoname,
        input_value=Literal("00111111"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_unique_value_iri(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/identical_values_LinkValue> ;
        sh:resultMessage "A resource may not have the same property and value more than one time." ;
        sh:resultPath onto:testHasLinkTo ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraint _:1 ;
        sh:sourceConstraintComponent sh:SPARQLConstraintComponent ;
        sh:sourceShape onto:ClassWithEverything_Unique ;
        sh:value <http://data/link_valueTarget_id> ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
        <http://data/identical_values_LinkValue> a onto:ClassWithEverything .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.SPARQLConstraintComponent,
        focus_node_iri=DATA.identical_values_LinkValue,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testHasLinkTo,
        severity=SH.Violation,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_unique_value_iri() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.UNIQUE_VALUE,
        res_iri=DATA.identical_values_LinkValue,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testHasLinkTo,
        input_value=DATA.link_valueTarget_id,
        severity=SH.Violation,
    )


@pytest.fixture
def report_coexist_with(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/missing_seqnum> ;
        sh:resultMessage "The property seqnum must be used together with isPartOf" ;
        sh:resultPath <http://api.knora.org/ontology/knora-api/v2#seqnum> ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent <http://datashapes.org/dash#CoExistsWithConstraintComponent> ;
        sh:sourceShape <http://api.knora.org/ontology/knora-api/shapes/v2#seqnum_PropShape> ] .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/missing_seqnum> a in-built:TestStillImageRepresentationWithSeqnum ;
        rdfs:label "Image with sequence"^^xsd:string ;
        knora-api:hasStillImageFileValue <http://data/file_value> ;
        knora-api:isPartOf <http://data/is_part_of_value> .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=DASH.CoExistsWithConstraintComponent,
        focus_node_iri=DATA.missing_seqnum,
        focus_node_type=IN_BUILT_ONTO.TestStillImageRepresentationWithSeqnum,
        result_path=KNORA_API.seqnum,
        severity=SH.Violation,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_coexist_with() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.SEQNUM_IS_PART_OF,
        res_iri=DATA.missing_seqnum,
        res_class=IN_BUILT_ONTO.TestStillImageRepresentationWithSeqnum,
        message=Literal("Coexist message from knora-api turtle"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_unknown_list_node(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ 
    a sh:ValidationResult ;
sh:detail _:bn_list_node_non_existent ;
sh:focusNode <http://data/list_node_non_existent> ;
sh:resultMessage "Value does not have shape <http://api.knora.org/ontology/knora-api/shapes/v2#firstList_NodeShape>" ;
sh:resultPath onto:testListProp ;
sh:resultSeverity sh:Violation ;
sh:sourceConstraintComponent sh:NodeConstraintComponent ;
sh:sourceShape onto:testListProp_PropShape ;
sh:value <http://data/value_list_node_non_existent> ] .
    
    _:bn_list_node_non_existent a sh:ValidationResult ;
    sh:focusNode <http://data/value_list_node_non_existent> ;
    sh:resultMessage "A valid node from the list 'firstList' must be used with this property." ;
    sh:resultPath knora-api:listValueAsListNode ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent sh:InConstraintComponent ;
    sh:sourceShape [ ] ;
    sh:value "firstList / other" .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/list_node_non_existent> a onto:ClassWithEverything .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    detail_bn = next(validation_g.objects(val_bn, SH.detail))
    detail = DetailBaseInfo(
        detail_bn=detail_bn,
        source_constraint_component=SH.InConstraintComponent,
    )
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.list_node_non_existent,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testListProp,
        severity=SH.Violation,
        detail=detail,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_unknown_list_node() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.GENERIC,
        res_iri=DATA.list_node_non_existent,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testListProp,
        message=Literal("A valid node from the list 'firstList' must be used with this property."),
        input_value=Literal("firstList / other"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_unknown_list_name(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
[
    a sh:ValidationResult ;
    sh:detail _:bn_list_name_non_existent ;
    sh:focusNode <http://data/list_name_non_existent> ;
    sh:resultMessage "Value does not have shape <http://rdfh.ch/lists/9999/b7p3ucDWQ5CZuKpVo-im7Q>" ;
    sh:resultPath <http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp> ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent sh:NodeConstraintComponent ;
    sh:sourceShape <http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp_PropShape> ;
    sh:value <http://data/value_list_name_non_existent> ] .

_:bn_list_name_non_existent a sh:ValidationResult ;
    sh:focusNode <http://data/value_list_name_non_existent> ;
    sh:resultMessage "A valid node from the list 'firstList' must be used with this property." ;
    sh:resultPath <http://api.knora.org/ontology/knora-api/v2#listValueAsListNode> ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent sh:InConstraintComponent ;
    sh:sourceShape _:bn_source ;
    sh:value "other / n1" .
    """
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    data_str = f"""{PREFIXES}
    <http://data/list_name_non_existent> a onto:ClassWithEverything .
    """
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    detail_bn = next(validation_g.objects(val_bn, SH.detail))
    detail = DetailBaseInfo(
        detail_bn=detail_bn,
        source_constraint_component=SH.InConstraintComponent,
    )
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.list_name_non_existent,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testListProp,
        severity=SH.Violation,
        detail=detail,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_unknown_list_name() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.GENERIC,
        res_iri=DATA.list_name_non_existent,
        res_class=ONTO.ClassWithEverything,
        property=ONTO.testListProp,
        message=Literal("A valid node from the list 'firstList' must be used with this property."),
        input_value=Literal("other / n1"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_missing_file_value(onto_graph: Graph) -> tuple[Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
            sh:focusNode <http://data/id_video_missing> ;
            sh:resultMessage "Cardinality 1" ;
            sh:resultPath <http://api.knora.org/ontology/knora-api/v2#hasMovingImageFileValue> ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
            sh:sourceShape _:n9f446ddf698b43ee9bc70be5f606ec25b13 ] .
    """
    data_str = f"""{PREFIXES}
    <http://data/id_video_missing> a <http://0.0.0.0:3333/ontology/9999/onto/v2#TestMovingImageRepresentation> ;
        rdfs:label "TestMovingImageRepresentation"^^xsd:string .
    """
    graphs = Graph()
    graphs.parse(data=validation_str, format="ttl")
    graphs.parse(data=data_str, format="ttl")
    graphs += onto_graph
    val_bn = next(graphs.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.MinCountConstraintComponent,
        focus_node_iri=DATA.id_video_missing,
        focus_node_type=ONTO.TestMovingImageRepresentation,
        result_path=KNORA_API.hasMovingImageFileValue,
        severity=SH.Violation,
    )
    return graphs, base_info


@pytest.fixture
def extracted_missing_file_value() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.MIN_CARD,
        res_iri=DATA.id_video_missing,
        res_class=ONTO.TestMovingImageRepresentation,
        property=KNORA_API.hasMovingImageFileValue,
        expected=Literal("Cardinality 1"),
        severity=SH.Violation,
    )


@pytest.fixture
def report_single_line_constraint_component(onto_graph: Graph) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f'''{PREFIXES}
        [ a sh:ValidationResult ;
            sh:focusNode <http://data/value_copyright_holder_with_newline> ;
            sh:resultMessage "The copyright holder must be a string without newlines." ;
            sh:resultPath <http://api.knora.org/ontology/knora-api/v2#hasCopyrightHolder> ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraintComponent <http://datashapes.org/dash#SingleLineConstraintComponent> ;
            sh:sourceShape [ ] ;
            sh:value """FirstLine
Second Line""" ] .
    '''
    data_str = f'''{PREFIXES}
<http://data/copyright_holder_with_newline> a <http://0.0.0.0:3333/ontology/9999/onto/v2#TestArchiveRepresentation> ;
rdfs:label "TestArchiveRepresentation zip"^^xsd:string ;
knora-api:hasArchiveFileValue <http://data/value_copyright_holder_with_newline> .

<http://data/value_copyright_holder_with_newline> a knora-api:ArchiveFileValue ;
    knora-api:fileValueHasFilename "this/is/filepath/file.zip"^^xsd:string ;
    knora-api:hasAuthorship "authorship_1"^^xsd:string ;
    knora-api:hasCopyrightHolder """FirstLine
Second Line"""^^xsd:string ;
    knora-api:hasLicense <http://rdfh.ch/licenses/cc-by-4.0> .
    '''
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=DASH.SingleLineConstraintComponent,
        focus_node_iri=DATA.copyright_holder_with_newline,
        focus_node_type=ONTO.TestArchiveRepresentation,
        result_path=KNORA_API.hasCopyrightHolder,
        severity=SH.Violation,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def report_single_line_constraint_component_content_is_value(
    onto_graph: Graph,
) -> tuple[Graph, Graph, ValidationResultBaseInfo]:
    validation_str = f'''{PREFIXES}
[ a sh:ValidationResult ;
            sh:detail _:detail_bn ;
            sh:focusNode <http://data/simple_text_with_newlines> ;
            sh:resultMessage "This property requires a TextValue" ;
            sh:resultPath <http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText> ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraintComponent sh:NodeConstraintComponent ;
            sh:sourceShape <http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText_PropShape> ;
            sh:value <http://data/value_simple_text_with_newlines> ].
            
    _:detail_bn a sh:ValidationResult ;
    sh:focusNode <http://data/value_simple_text_with_newlines> ;
    sh:resultMessage "The value must be a non-empty string without newlines." ;
    sh:resultPath <http://api.knora.org/ontology/knora-api/v2#valueAsString> ;
    sh:resultSeverity sh:Violation ;
    sh:sourceConstraintComponent <http://datashapes.org/dash#SingleLineConstraintComponent> ;
    sh:sourceShape [ ] ;
    sh:value """This may not

have newlines""" .
    '''
    data_str = f'''{PREFIXES}
<http://data/simple_text_with_newlines> a onto:ClassWithEverything ;
    rdfs:label "With linebreaks"^^xsd:string ;
    onto:testSimpleText <http://data/value_simple_text_with_newlines> .

<http://data/value_simple_text_with_newlines> a knora-api:TextValue ;
    knora-api:valueAsString """This may not

have newlines"""^^xsd:string .
    '''
    validation_g = Graph()
    validation_g.parse(data=validation_str, format="ttl")
    onto_data_g = Graph()
    onto_data_g += onto_graph
    onto_data_g.parse(data=data_str, format="ttl")
    val_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
    detail_bn = next(validation_g.objects(val_bn, SH.detail))
    detail = DetailBaseInfo(
        detail_bn=detail_bn,
        source_constraint_component=DASH.SingleLineConstraintComponent,
    )
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.NodeConstraintComponent,
        focus_node_iri=DATA.simple_text_with_newlines,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=ONTO.testSimpleText,
        severity=SH.Violation,
        detail=detail,
    )
    return validation_g, onto_data_g, base_info


@pytest.fixture
def extracted_single_line_constraint_component() -> ValidationResult:
    return ValidationResult(
        violation_type=ViolationType.GENERIC,
        res_iri=DATA.copyright_holder_with_newline,
        res_class=ONTO.TestArchiveRepresentation,
        property=KNORA_API.hasCopyrightHolder,
        message=Literal("The copyright holder must be a string without newlines."),
        input_value=Literal("with newline"),
        severity=SH.Violation,
    )


@pytest.fixture
def result_unknown_component(onto_graph: Graph) -> tuple[Graph, ValidationResultBaseInfo]:
    validation_str = f"""{PREFIXES}
    [ a sh:ValidationResult ;
        sh:focusNode <http://data/empty_label> ;
        sh:resultMessage "The label must be a non-empty string" ;
        sh:resultPath rdfs:label ;
        sh:resultSeverity sh:Violation ;
        sh:sourceConstraintComponent sh:UniqueLangConstraintComponent ;
        sh:sourceShape api-shapes:rdfsLabel_Shape ;
        sh:value " " ] .
    """
    data_str = f"""{PREFIXES}
        <http://data/empty_label> a onto:ClassWithEverything ;
            rdfs:label " "^^xsd:string .
    """
    graphs = Graph()
    graphs.parse(data=validation_str, format="ttl")
    graphs.parse(data=data_str, format="ttl")
    graphs += onto_graph
    val_bn = next(graphs.subjects(RDF.type, SH.ValidationResult))
    base_info = ValidationResultBaseInfo(
        result_bn=val_bn,
        source_constraint_component=SH.PatternConstraintComponent,
        focus_node_iri=DATA.empty_label,
        focus_node_type=ONTO.ClassWithEverything,
        result_path=RDFS.label,
        severity=SH.Violation,
    )
    return graphs, base_info



from __future__ import annotations

import datetime
import json
import uuid
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import regex
from lxml import etree
from regex import Match

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.error.xmllib_warnings_util import raise_input_error
from dsp_tools.xmllib.internal.checkers import is_nonempty_value_internal
from dsp_tools.xmllib.internal.constants import KNOWN_XML_TAG_REGEXES
from dsp_tools.xmllib.internal.input_converters import unescape_reserved_xml_chars
from dsp_tools.xmllib.models.config_options import NewlineReplacement
from dsp_tools.xmllib.models.licenses.other import LicenseOther
from dsp_tools.xmllib.models.licenses.recommended import License
from dsp_tools.xmllib.models.licenses.recommended import LicenseRecommended
from dsp_tools.xmllib.value_converters import replace_newlines_with_tags


def create_footnote_string(
    footnote_text: str, newline_replacement_option: NewlineReplacement = NewlineReplacement.LINEBREAK
) -> str:
    """
    Takes the text for a footnote, and returns a string with the correct formatting.
    You can use this if you want to add the footnote to a string.
    Currently, the newline replacement options are restricted to `LINEBREAK` and `NONE`.
    The reserved characters `<`, `>` and `&` will be escaped temporarily,
    but they will be correctly displayed in DSP-APP.

    Attention:
        - The text in the footnote may be richtext, i.e. contain XML tags.
        - Not all tags supported in ordinary richtext are currently implemented.
        - The allowed tags are:
            - `<br>` (break line)
            - `<strong>` (bold)
            - `<em>` (italic)
            - `<u>` (underline)
            - `<strike>` (strike through)
            - `<a href="URI">` (link to a URI)
            - `<a class="salsah-link" href="Knora IRI">` (link to a resource)

    Args:
        footnote_text: Text for the footnote
        newline_replacement_option: options to replace newlines

    Raises:
        InputError: If the text is empty, or if a newline replacement which is not implemented is entered

    Returns:
        The footnote as a string

    Examples:
        ```python
        result = xmllib.create_footnote_string("Text")
        # result == '<footnote content="Text"/>'
        ```

        ```python
        result = xmllib.create_footnote_string("Text\\nSecond Line")
        # result == '<footnote content="Text&lt;br/&gt;Second Line"/>'
        ```

        ```python
        result = xmllib.create_footnote_string("Already escaped &lt;&gt;")
        # already escaped characters will not be escaped again
        # result == '<footnote content="Already escaped &lt;&gt;"/>'
        ```
    """
    text_tag = create_footnote_element(footnote_text, newline_replacement_option)
    return etree.tostring(text_tag, encoding="unicode")


def create_footnote_element(
    footnote_text: str, newline_replacement_option: NewlineReplacement = NewlineReplacement.LINEBREAK
) -> etree._Element:
    """
    Takes the text for a footnote, and returns an `etree.Element`.
    You can use this if you are working with `lxml`.
    Currently, the newline replacement options are restricted to `LINEBREAK` and `NONE`.

    Attention:
        - The text in the footnote may be richtext, i.e. contain XML tags.
        - Not all tags supported in ordinary richtext are currently implemented.
        - The allowed tags are:
            - `<br>` (break line)
            - `<strong>` (bold)
            - `<em>` (italic)
            - `<u>` (underline)
            - `<strike>` (strike through)
            - `<a href="URI">` (link to a URI)
            - `<a class="salsah-link" href="Knora IRI">` (link to a resource)

    Args:
        footnote_text: Text for the footnote
        newline_replacement_option: options to replace newlines

    Raises:
        InputError: If the text is empty, or if a newline replacement which is not implemented is entered

    Returns:
        The footnote as a string
    """
    if newline_replacement_option not in {NewlineReplacement.LINEBREAK, NewlineReplacement.NONE}:
        raise_input_error(MessageInfo("Currently the only supported newline replacement is linebreak (<br/>) or None."))
    if not is_nonempty_value_internal(footnote_text):
        raise_input_error(MessageInfo("The input value is empty."))
    footnote_text = replace_newlines_with_tags(str(footnote_text), newline_replacement_option)
    unescaped_text = unescape_reserved_xml_chars(footnote_text)
    return etree.Element("footnote", attrib={"content": unescaped_text})


def create_standoff_link_to_resource(resource_id: str, displayed_text: str) -> str:
    """
    Creates a standoff link to a resource.

    Args:
        resource_id: ID of the resource that is linked
        displayed_text: text to display for the embedded link

    Returns:
        A standoff link in string form.

    Raises:
        InputError: if the resource ID or the displayed text are empty

    Examples:
        ```python
        result = xmllib.create_standoff_link_to_resource("resource_id", "Text")
        # result == '<a class="salsah-link" href="IRI:resource_id:IRI">Text</a>'
        ```
    """
    if not all([is_nonempty_value_internal(resource_id), is_nonempty_value_internal(displayed_text)]):
        msg_str = (
            f"The entered resource ID and displayed text may not be empty. "
            f"Your input: resource_id '{resource_id}' / displayed_text '{displayed_text}'"
        )
        raise_input_error(MessageInfo(msg_str))
    attribs = {"class": "salsah-link", "href": f"IRI:{resource_id}:IRI"}
    ele = etree.Element("a", attrib=attribs)
    ele.text = displayed_text
    return etree.tostring(ele, encoding="unicode")


def create_standoff_link_to_uri(uri: str, displayed_text: str) -> str:
    """
    Creates a standoff link to a URI.

    Args:
        uri: the target URI that should be linked to
        displayed_text: text to display for the embedded link

    Returns:
        A standoff link in string form.

    Raises:
        InputError: if the URI or the displayed text are empty

    Examples:
        ```python
        result = xmllib.create_standoff_link_to_uri("https://www.dasch.swiss/", "This is DaSCH")
        # result == '<a href="https://www.dasch.swiss/">This is DaSCH</a>'
        ```
    """
    if not all([is_nonempty_value_internal(uri), is_nonempty_value_internal(displayed_text)]):
        msg_str = (
            f"The entered URI and displayed text may not be empty. "
            f"Your input: uri '{uri}' / displayed_text '{displayed_text}'"
        )
        raise_input_error(MessageInfo(msg_str))
    attribs = {"href": uri}
    ele = etree.Element("a", attrib=attribs)
    ele.text = displayed_text
    return etree.tostring(ele, encoding="unicode")


def _get_label_to_node_one_list(
    list_section: list[dict[str, Any]], list_name: str, language_of_label: str
) -> dict[str, str]:
    json_subset = [x for x in list_section if x["name"] == list_name]
    # json_subset is a list containing one item, namely the json object containing the entire json-list
    res = {}
    for label, name in _name_label_mapper_iterator(json_subset, language_of_label):
        if name != list_name:
            res[label] = name
            res[label.strip().lower()] = name
    return res


def _get_label_to_node_all_lists(
    list_section: list[dict[str, Any]], language_of_label: str
) -> dict[str, dict[str, str]]:
    mapper = {}
    for li in list_section:
        mapper[li["name"]] = _get_label_to_node_one_list(list_section, li["name"], language_of_label)
    return mapper


def _get_property_to_list_name_mapping(ontologies: list[dict[str, Any]], default_ontology: str) -> dict[str, str]:
    prop_lookup = {}
    for onto in ontologies:
        prefix = onto["name"]
        property_section = onto["properties"]
        for prop in property_section:
            if prop["gui_element"] == "List":
                prefixed_prop = f"{prefix}:{prop['name']}"
                prop_lookup[prefixed_prop] = prop["gui_attributes"]["hlist"]
    default_props = {
        k.replace(default_ontology, "", 1): v for k, v in prop_lookup.items() if k.startswith(f"{default_ontology}:")
    }
    prop_lookup = prop_lookup | default_props
    return prop_lookup


@dataclass
class ListLookup:
    _lookup: dict[str, dict[str, str]]
    _prop_to_list_name: dict[str, str]
    _label_language: str

    @staticmethod
    def create_new(project_json_path: str | Path, language_of_label: str, default_ontology: str) -> ListLookup:
        """
        Creates a list lookup based on list labels in a specified language and returning list node names.
        Works for all lists in a project.json

        Args:
            project_json_path: path to a JSON project file (a.k.a. ontology)
            language_of_label: label language used for the list
            default_ontology: ontology prefix which is defined as default in the XML file

        Returns:
            `ListLookup` for a project

        Examples:
            ```python
            list_lookup = xmllib.ListLookup.create_new(
                project_json_path="project.json",
                language_of_label="en",
                default_ontology="default-onto",
            )
            ```
        """
        with open(project_json_path, encoding="utf-8") as f:
            json_file = json.load(f)
        label_to_list_node_lookup = _get_label_to_node_all_lists(json_file["project"]["lists"], language_of_label)
        prop_to_list_mapper = _get_property_to_list_name_mapping(json_file["project"]["ontologies"], default_ontology)
        return ListLookup(
            _lookup=label_to_list_node_lookup,
            _prop_to_list_name=prop_to_list_mapper,
            _label_language=language_of_label,
        )

    def get_node_via_list_name(self, list_name: str, node_label: str) -> str:
        """
        Returns the list node name based on a label.
        The language of the label was specified when creating the `ListLookup`.

        Args:
            list_name: name of the list
            node_label: label of the node

        Returns:
            node name

        Examples:
            ```python
            node_name = list_lookup.get_node_via_list_name(
                list_name="list1",
                node_label="Label 1"  # or: "label 1" (capitalisation is not relevant)
            )
            # node_name == "node1"
            ```
        """
        if not (list_lookup := self._lookup.get(list_name)):
            emit_xmllib_input_warning(
                MessageInfo(f"The entered list name '{list_name}' was not found. An empty string is returned.")
            )
            return ""
        if not (found_node := list_lookup.get(node_label)):
            emit_xmllib_input_warning(
                MessageInfo(
                    f"'{node_label}' was not recognised as label of the list '{list_name}'. "
                    f"This ListLookup is configured for '{self._label_language}' labels. An empty string is returned."
                )
            )
            return ""
        return found_node

    def get_list_name_and_node_via_property(self, prop_name: str, node_label: str) -> tuple[str, str]:
        """
        Returns the list name and the node name based on a property that is used with the list and the label of a node.
        The language of the label was specified when creating the `ListLookup`.
        The list name needs to be referenced in the XML file.

        Args:
            prop_name: name of the list
            node_label: label of the node

        Returns:
            list name and node name

        Examples:
            ```python
            list_name, node_name = list_lookup.get_list_name_and_node_via_property(
                prop_name=":hasList",  # or: "default-onto:hasList"
                node_label="label 1"
            )
            # list_name == "list1"
            # node_name == "node1"
            ```
        """
        if not (list_name := self.get_list_name_via_property(prop_name)):
            return "", ""
        return list_name, self.get_node_via_list_name(list_name, node_label)

    def get_list_name_via_property(self, prop_name: str) -> str:
        """
        Returns the list name as specified in the ontology for a property.
        The list name needs to be referenced in the XML file.

        Args:
            prop_name: name of the property

        Returns:
            Name of the list

        Examples:
            ```python
            list_name = list_lookup.get_list_name_via_property(
                prop_name=":hasList",  # or: "default-onto:hasList"
            )
            # list_name == "list1"
            ```
        """
        if not (list_name := self._prop_to_list_name.get(prop_name)):
            emit_xmllib_input_warning(
                MessageInfo(f"The entered property '{prop_name}' was not found. An empty string is returned.")
            )
            return ""
        return list_name


def get_list_nodes_from_string_via_list_name(
    string_with_list_labels: str, label_separator: str, list_name: str, list_lookup: ListLookup
) -> list[str]:
    """
    Takes a string containing list labels, the separator by which they can be split,
    a property name and the list lookup.
    Resolves the labels and returns the list name to be referenced in the XML file and a list of node names.
    If the string is empty, it returns an empty list.

    Args:
        string_with_list_labels: the string containing the labels
        label_separator: separator in the string that contains the labels
        list_name: name of the list
        list_lookup: `ListLookup` of the project

    Returns:
        The name of the list and a list of node names.

    Examples:
        ```python
        string_with_list_labels = "Label 1; Label 2"
        nodes = xmllib.get_list_nodes_from_string_via_list_name(
            string_with_list_labels=string_with_list_labels,
            label_separator=";",
            list_name="list1",
            list_lookup=list_lookup,
        )
        # nodes == ["node1", "node2"]
        ```

        ```python
        string_with_list_labels = ""
        nodes = xmllib.get_list_nodes_from_string_via_list_name(
            string_with_list_labels=string_with_list_labels,
            label_separator=";",
            list_name="list1",
            list_lookup=list_lookup,
        )
        # nodes == []
        ```

        ```python
        string_with_list_labels = pd.NA
        nodes = xmllib.get_list_nodes_from_string_via_list_name(
            string_with_list_labels=string_with_list_labels,
            label_separator=";",
            list_name="list1",
            list_lookup=list_lookup,
        )
        # nodes == []
        ```
    """
    if not is_nonempty_value_internal(string_with_list_labels):
        return []
    labels_list = create_list_from_string(string_with_list_labels, label_separator)
    nodes_list = [list_lookup.get_node_via_list_name(list_name, label) for label in labels_list]
    return nodes_list


def get_list_nodes_from_string_via_property(
    string_with_list_labels: str, label_separator: str, property_name: str, list_lookup: ListLookup
) -> tuple[str, list[str]]:
    """
    Takes a string containing list labels, the separator by which they can be split,
    a property name and the list lookup.
    Resolves the labels and returns the list name to be referenced in the XML file and a list of node names.
    If the string is empty, it returns an empty list.

    Args:
        string_with_list_labels: the string containing the labels
        label_separator: separator in the string that contains the labels
        property_name: name of the property
        list_lookup: `ListLookup` of the project

    Returns:
        The name of the list and a list of node names.

    Examples:
        ```python
        string_with_list_labels = "Label 1; Label 2"
        list_name, nodes = xmllib.get_list_nodes_from_string_via_property(
            string_with_list_labels=string_with_list_labels,
            label_separator=";",
            property_name=":hasList",
            list_lookup=list_lookup,
        )
        # list_name == "list1"
        # nodes == ["node1", "node2"]
        ```

        ```python
        string_with_list_labels = ""
        list_name, nodes = xmllib.get_list_nodes_from_string_via_property(
            string_with_list_labels=string_with_list_labels,
            label_separator=";",
            property_name=":hasList",
            list_lookup=list_lookup,
        )
        # list_name == ""
        # nodes == []
        ```

        ```python
        string_with_list_labels = pd.NA
        list_name, nodes = xmllib.get_list_nodes_from_string_via_property(
            string_with_list_labels=string_with_list_labels,
            label_separator=";",
            property_name=":hasList",
            list_lookup=list_lookup,
        )
        # list_name == ""
        # nodes == []
        ```
    """
    if not is_nonempty_value_internal(string_with_list_labels):
        return "", []
    labels_list = create_list_from_string(string_with_list_labels, label_separator)
    list_name = ""
    nodes = []
    for lbl in labels_list:
        list_name, node_name = list_lookup.get_list_name_and_node_via_property(property_name, lbl)
        nodes.append(node_name)
    return list_name, nodes


def _name_label_mapper_iterator(
    json_subset: list[dict[str, Any]],
    language_of_label: str,
) -> Iterable[tuple[str, str]]:
    """
    Go through list nodes of a JSON project and yield (label, name) pairs.

    Args:
        json_subset: list of DSP lists (a DSP list being a dictionary with the keys "name", "labels" and "nodes")
        language_of_label: which language of the label to choose

    Yields:
        (label, name) pairs
    """
    for node in json_subset:
        # node is the json object containing the entire json-list
        if "nodes" in node:
            # "nodes" is the json sub-object containing the entries of the json-list
            yield from _name_label_mapper_iterator(node["nodes"], language_of_label)
            # each yielded value is a (label, name) pair of a single list entry
        if "name" in node:
            yield node["labels"][language_of_label], node["name"]
            # the actual values of the name and the label


def escape_reserved_xml_characters(text: str) -> str:
    """
    From richtext strings (encoding="xml"), escape the reserved characters `<`, `>` and `&`,
    but only if they are not part of a standard standoff tag or escape sequence.

    [See the documentation for the standard standoff tags allowed by DSP-API,
    which will not be escaped.](https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/)

    Args:
        text: the richtext string to be escaped

    Returns:
        The escaped richtext string

    Examples:
        ```python
        result = xmllib.escape_reserved_xml_characters("Text <unknownTag>")
        # result == "Text &lt;unknownTag&gt;"
        ```

        ```python
        result = xmllib.escape_reserved_xml_characters("Text <br/> text after")
        # result == "Text <br/> text after"
        ```
    """
    allowed_tags_regex = "|".join(KNOWN_XML_TAG_REGEXES)
    lookahead = rf"(?!/?({allowed_tags_regex})/?>)"
    illegal_lt = rf"<{lookahead}"
    lookbehind = rf"(?<!</?({allowed_tags_regex})/?)"
    illegal_gt = rf"{lookbehind}>"
    illegal_amp = r"&(?![#a-zA-Z0-9]+;)"
    text = regex.sub(illegal_lt, "&lt;", text)
    text = regex.sub(illegal_gt, "&gt;", text)
    text = regex.sub(illegal_amp, "&amp;", text)
    return text


def find_dates_in_string(string: str) -> set[str]:
    """
    Checks if a string contains date values (single dates, or date ranges),
    and return all found dates as set of DSP-formatted strings.
    Returns an empty set if no date was found.
    [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#date).

    Notes:
        - If no era or calendar is given, dates are interpreted in the Common Era and the Gregorian calendar.
        - Standalone numbers from 000-2999, in 3/4-digit form, are interpreted as years CE.
        - If a number (with any number of digits) is followed by CE, C.E., AD, A.D., it is interpreted as years CE.
        - If a number (with any number of digits) is followed by BCE, BC, B.C., B.C.E., av. J.-C.,
          it is interpreted as years BCE.
        - Dates written with slashes are always interpreted in a European manner: 5/11/2021 is the 5th of November.
        - In the European notation, 2-digit years are expanded to 4 digits, with the current year as watershed:
            - 30.4.24 -> 30.04.2024
            - 30.4.50 -> 30.04.1950

    Currently supported date formats:
        - 0476-09-04 -> GREGORIAN:CE:0476-09-04:CE:0476-09-04
        - 0476_09_04 -> GREGORIAN:CE:0476-09-04:CE:0476-09-04
        - 30.4.2021 -> GREGORIAN:CE:2021-04-30:CE:2021-04-30
        - 30.4.21 -> GREGORIAN:CE:2021-04-30:CE:2021-04-30
        - 5/11/2021 -> GREGORIAN:CE:2021-11-05:CE:2021-11-05
        - Jan 26, 1993 -> GREGORIAN:CE:1993-01-26:CE:1993-01-26
        - 26 Jan 1993 -> GREGORIAN:CE:1993-01-26:CE:1993-01-26
        - 26 January 1993 -> GREGORIAN:CE:1993-01-26:CE:1993-01-26
        - 26. Jan. 1993 -> GREGORIAN:CE:1993-01-26:CE:1993-01-26
        - 26. Januar 1993 -> GREGORIAN:CE:1993-01-26:CE:1993-01-26
        - 28.2.-1.12.1515 -> GREGORIAN:CE:1515-02-28:CE:1515-12-01
        - 25.-26.2.0800 -> GREGORIAN:CE:0800-02-25:CE:0800-02-26
        - 1.9.2022-3.1.2024 -> GREGORIAN:CE:2022-09-01:CE:2024-01-03
        - 1848 -> GREGORIAN:CE:1848:CE:1848
        - 1849/1850 -> GREGORIAN:CE:1849:CE:1850
        - 1849/50 -> GREGORIAN:CE:1849:CE:1850
        - 1845-50 -> GREGORIAN:CE:1845:CE:1850
        - 840-50 -> GREGORIAN:CE:840:CE:850
        - 840-1 -> GREGORIAN:CE:840:CE:841
        - 9 BC / 9 B.C. / 9 B.C.E. / 9 BCE -> GREGORIAN:BC:9:BC:9
        - 20 BCE - 50 CE -> GREGORIAN:BC:20:CE:50
        - 1000-900 av. J.-C. -> GREGORIAN:BC:1000:BC:900
        - 45 av. J.-C. -> GREGORIAN:BC:45:BC:45

    Args:
        string: string to check

    Returns:
        (possibly empty) set of DSP-formatted date strings

    Examples:
        ```python
        result = xmllib.find_dates_in_string("1849/1850")
        # result == {"GREGORIAN:CE:1849:CE:1850"}
        ```

        ```python
        result = xmllib.find_dates_in_string("not a valid date")
        # result == {}
        ```

        ```python
        result = xmllib.find_dates_in_string("first date: 2024. Second: 2025.")
        # result == {"GREGORIAN:CE:2024:CE:2024", "GREGORIAN:CE:2025:CE:2025"}
        ```
    """

    # sanitise input, just in case that the function was called on an empty or N/A cell
    if not is_nonempty_value_internal(string):
        return set()
    return _find_dates_in_string(string)


_months_dict = {
    "January": 1,
    "Januar": 1,
    "Jan": 1,
    "February": 2,
    "Februar": 2,
    "Feb": 2,
    "March": 3,
    "März": 3,
    "Mar": 3,
    "April": 4,
    "Apr": 4,
    "May": 5,
    "Mai": 5,
    "June": 6,
    "Juni": 6,
    "Jun": 6,
    "July": 7,
    "Juli": 7,
    "Jul": 7,
    "August": 8,
    "Aug": 8,
    "September": 9,
    "Sept": 9,
    "October": 10,
    "Oktober": 10,
    "Oct": 10,
    "Okt": 10,
    "November": 11,
    "Nov": 11,
    "December": 12,
    "Dezember": 12,
    "Dec": 12,
    "Dez": 12,
}
all_months = "|".join(_months_dict)


def _find_dates_in_string(string: str) -> set[str]:
    year_regex = r"([0-2]?[0-9][0-9][0-9])"
    year_regex_2_or_4_digits = r"((?:[0-2]?[0-9])?[0-9][0-9])"
    month_regex = r"([0-1]?[0-9])"
    day_regex = r"([0-3]?[0-9])"
    sep_regex = r"[\./]"
    lookbehind = r"(?<![0-9A-Za-z])"
    lookahead = r"(?![0-9A-Za-z])"
    range_operator_regex = r" ?- ?"

    remaining_string = string
    results: set[str | None] = set()

    remaining_string = _extract_already_parsed_date(remaining_string, results)

    remaining_string = _find_english_BC_or_CE_dates(
        string=remaining_string,
        lookbehind=lookbehind,
        lookahead=lookahead,
        range_operator_regex=range_operator_regex,
        results=results,
    )

    remaining_string = _find_french_bc_dates(
        string=remaining_string,
        lookbehind=lookbehind,
        lookahead=lookahead,
        range_operator_regex=range_operator_regex,
        results=results,
    )

    # template: 2021-01-01 | 2015_01_02
    iso_dates_regex = rf"{lookbehind}{year_regex}[_-]([0-1][0-9])[_-]([0-3][0-9]){lookahead}"
    if iso_dates := list(regex.finditer(iso_dates_regex, remaining_string)):
        results.update(_from_iso_date(x) for x in iso_dates)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in iso_dates])

    # template: 6.-8.3.1948 | 6/2/1947 - 24.03.1948
    eur_date_range_regex = (
        rf"{lookbehind}"
        rf"{day_regex}{sep_regex}(?:{month_regex}{sep_regex}{year_regex_2_or_4_digits}?)?{range_operator_regex}"
        rf"{day_regex}{sep_regex}{month_regex}{sep_regex}{year_regex_2_or_4_digits}"
        rf"{lookahead}"
    )
    if eur_date_ranges := list(regex.finditer(eur_date_range_regex, remaining_string)):
        results.update(_from_eur_date_range(x) for x in eur_date_ranges)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in eur_date_ranges])

    # template: 1.4.2021 | 5/11/2021
    eur_date_regex = rf"{lookbehind}{day_regex}{sep_regex}{month_regex}{sep_regex}{year_regex_2_or_4_digits}{lookahead}"
    if eur_dates := list(regex.finditer(eur_date_regex, remaining_string)):
        results.update(_from_eur_date(x) for x in eur_dates)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in eur_dates])

    # template: March 9, 1908 | March5,1908 | May 11, 1906
    monthname_date_regex = rf"{lookbehind}({all_months}) ?{day_regex}, ?{year_regex}{lookahead}"
    if monthname_dates := list(regex.finditer(monthname_date_regex, remaining_string)):
        results.update(_from_monthname_date(x) for x in monthname_dates)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in monthname_dates])

    # template: 9 March 1908
    monthname_after_day_regex = rf"{lookbehind}{day_regex} ?({all_months}) ?{year_regex}{lookahead}"
    if monthname_after_days := list(regex.finditer(monthname_after_day_regex, remaining_string)):
        results.update(_from_monthname_after_day(x) for x in monthname_after_days)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in monthname_after_days])

    # template: 26. Januar 1993 | 26. Jan. 1993 | 26. Jan 1993
    german_monthname_date_regex = rf"{lookbehind}{day_regex}\.? ?({all_months})\.? ?{year_regex}{lookahead}"
    if german_monthname_dates := list(regex.finditer(german_monthname_date_regex, remaining_string)):
        results.update(_from_german_monthname_date(x) for x in german_monthname_dates)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in german_monthname_dates])

    # template: 1849/50 | 1849-50 | 1849/1850
    if year_ranges := list(regex.finditer(lookbehind + year_regex + r"[/-](\d{1,4})" + lookahead, remaining_string)):
        results.update(_from_year_range(x) for x in year_ranges)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in year_ranges])

    # template: 1907
    if year_onlies := list(regex.finditer(rf"{lookbehind}{year_regex}{lookahead}", remaining_string)):
        results.update(f"GREGORIAN:CE:{int(x.group(0))}:CE:{int(x.group(0))}" for x in year_onlies)
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in year_onlies])

    return {x for x in results if x}


def _remove_used_spans(string: str, spans: list[tuple[int, int]]) -> str:
    """Once a regex has matched parts of the original string, remove these parts, so that they're not matched again."""
    for start, end in reversed(spans):
        string = string[:start] + string[end:]
    return string


def _find_english_BC_or_CE_dates(
    string: str,
    lookbehind: str,
    lookahead: str,
    range_operator_regex: str,
    results: set[str | None],
) -> str:
    eraless_date_regex = r"(\d+)"
    bc_era_regex = r"(?:BC|BCE|B\.C\.|B\.C\.E\.)"
    bc_date_regex = rf"(?:{eraless_date_regex} ?{bc_era_regex})"
    ce_era_regex = r"(?:CE|AD|C\.E\.|A\.D\.)"
    ce_date_regex = rf"(?:{eraless_date_regex} ?{ce_era_regex})"
    bc_or_ce_date_regex = rf"(?:{bc_date_regex}|{ce_date_regex})"

    remaining_string = string
    results_new: set[str | None] = set()

    range_regex = (
        rf"{lookbehind}(?:{bc_or_ce_date_regex}|{eraless_date_regex})"
        rf"{range_operator_regex}"
        rf"{bc_or_ce_date_regex}{lookahead}"
    )
    if matchs := list(regex.finditer(range_regex, remaining_string)):
        results_new.update(
            _from_english_BC_or_CE_range(
                string=x.group(0),
                range_operator_regex=range_operator_regex,
                bc_era_regex=bc_era_regex,
                ce_era_regex=ce_era_regex,
                eraless_date_regex=eraless_date_regex,
            )
            for x in matchs
        )
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in matchs])

    if matchs := list(regex.finditer(rf"{lookbehind}{bc_date_regex}{lookahead}", remaining_string)):
        results_new.update({f"GREGORIAN:BC:{x.group(1)}:BC:{x.group(1)}" for x in matchs})
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in matchs])

    if matchs := list(regex.finditer(rf"{lookbehind}{ce_date_regex}{lookahead}", remaining_string)):
        results_new.update({f"GREGORIAN:CE:{x.group(1)}:CE:{x.group(1)}" for x in matchs})
        remaining_string = _remove_used_spans(remaining_string, [x.span() for x in matchs])

    results.update({x for x in results_new if x})
    return remaining_string


def _from_english_BC_or_CE_range(
    string: str, range_operator_regex: str, bc_era_regex: str, ce_era_regex: str, eraless_date_regex: str
) -> str | None:
    split_result = regex.split(range_operator_regex, string)
    if len(split_result) != 2:
        return None
    start_raw, end_raw = split_result
    if regex.search(bc_era_regex, end_raw):
        end_era = "BC"
    elif regex.search(ce_era_regex, end_raw):
        end_era = "CE"
    else:
        return None

    if regex.search(bc_era_regex, start_raw):
        start_era = "BC"
    elif regex.search(ce_era_regex, start_raw):
        start_era = "CE"
    else:
        start_era = end_era

    if not (start_year_match := regex.search(eraless_date_regex, start_raw)):
        return None
    if not (end_year_match := regex.search(eraless_date_regex, end_raw)):
        return None

    return f"GREGORIAN:{start_era}:{start_year_match.group(0)}:{end_era}:{end_year_match.group(0)}"


def _find_french_bc_dates(
    string: str,
    lookbehind: str,
    lookahead: str,
    range_operator_regex: str,
    results: set[str | None],
) -> str:
    remaining_string = string
    results_new: set[str | None] = set()
    french_bc_regex = r"av(?:\. |\.| )J\.?-?C\.?"

    year_regex = r"\d{1,5}"
    year_range_regex = rf"{lookbehind}({year_regex}){range_operator_regex}({year_regex}) {french_bc_regex}{lookahead}"
    for year_range in reversed(list(regex.finditer(year_range_regex, remaining_string))):
        start_year = int(year_range.group(1))
        end_year = int(year_range.group(2))
        if end_year > start_year:
            continue
        results_new.add(f"GREGORIAN:BC:{start_year}:BC:{end_year}")
        remaining_string = _remove_used_spans(remaining_string, [year_range.span()])

    single_year_regex = rf"{lookbehind}({year_regex}) {french_bc_regex}{lookahead}"
    for single_year in reversed(list(regex.finditer(single_year_regex, remaining_string))):
        start_year = int(single_year.group(1))
        results_new.add(f"GREGORIAN:BC:{start_year}:BC:{start_year}")
        remaining_string = _remove_used_spans(remaining_string, [single_year.span()])

    results.update({x for x in results_new if x})
    return remaining_string


def _from_iso_date(iso_date: Match[str]) -> str | None:
    year = int(iso_date.group(1))
    month = int(iso_date.group(2))
    day = int(iso_date.group(3))
    try:
        date = datetime.date(year, month, day)
        return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"
    except ValueError:
        return None


def _expand_2_digit_year(year: int) -> int:
    current_year = datetime.date.today().year - 2000
    if year <= current_year:
        return year + 2000
    elif year <= 99:
        return year + 1900
    else:
        return year


def _from_eur_date_range(eur_date_range: Match[str]) -> str | None:
    startday = int(eur_date_range.group(1))
    startmonth = int(eur_date_range.group(2)) if eur_date_range.group(2) else int(eur_date_range.group(5))
    startyear = int(eur_date_range.group(3)) if eur_date_range.group(3) else int(eur_date_range.group(6))
    startyear = _expand_2_digit_year(startyear)
    endday = int(eur_date_range.group(4))
    endmonth = int(eur_date_range.group(5))
    endyear = int(eur_date_range.group(6))
    endyear = _expand_2_digit_year(endyear)
    try:
        startdate = datetime.date(startyear, startmonth, startday)
        enddate = datetime.date(endyear, endmonth, endday)
    except ValueError:
        return None
    if enddate < startdate:
        return None
    return f"GREGORIAN:CE:{startdate.isoformat()}:CE:{enddate.isoformat()}"


def _from_eur_date(eur_date: Match[str]) -> str | None:
    startday = int(eur_date.group(1))
    startmonth = int(eur_date.group(2))
    startyear = int(eur_date.group(3))
    startyear = _expand_2_digit_year(startyear)
    try:
        date = datetime.date(startyear, startmonth, startday)
        return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"
    except ValueError:
        return None


def _from_monthname_date(monthname_date: Match[str]) -> str | None:
    day = int(monthname_date.group(2))
    month = _months_dict[monthname_date.group(1)]
    year = int(monthname_date.group(3))
    try:
        date = datetime.date(year, month, day)
        return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"
    except ValueError:
        return None


def _from_monthname_after_day(monthname_after_day: Match[str]) -> str | None:
    day = int(monthname_after_day.group(1))
    month = _months_dict[monthname_after_day.group(2)]
    year = int(monthname_after_day.group(3))
    try:
        date = datetime.date(year, month, day)
        return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"
    except ValueError:
        return None


def _from_german_monthname_date(german_monthname_date: Match[str]) -> str | None:
    day = int(german_monthname_date.group(1))
    month = _months_dict[german_monthname_date.group(2)]
    year = int(german_monthname_date.group(3))
    try:
        date = datetime.date(year, month, day)
        return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"
    except ValueError:
        return None


def _from_year_range(year_range: Match[str]) -> str | None:
    startyear = int(year_range.group(1))
    endyear = int(year_range.group(2))
    if endyear // 10 == 0:
        # endyear is only 1-digit: add the first 2-3 digits of startyear
        endyear = startyear // 10 * 10 + endyear
    elif endyear // 100 == 0:
        # endyear is only 2-digit: add the first 1-2 digits of startyear
        endyear = startyear // 100 * 100 + endyear
    if endyear <= startyear:
        return None
    return f"GREGORIAN:CE:{startyear}:CE:{endyear}"


def _extract_already_parsed_date(string: str, results: set[str | None]) -> str:
    rgx_year = r"\d+(-\d{2}(-\d{2})?)?"
    era_with_colon = r"(CE:|BC:)"
    rgx = rf"(GREGORIAN|JULIAN|ISLAMIC):{era_with_colon}{rgx_year}:{era_with_colon}?{rgx_year}"
    if matchs := list(regex.finditer(rgx, string)):
        results.update({x.group(0) for x in matchs})
        remaining_string = _remove_used_spans(string, [x.span() for x in matchs])
        return remaining_string
    return string


def make_xsd_compatible_id(input_value: str | float | int) -> str:
    """
    An xsd:ID may not contain all types of special characters,
    and it must start with a letter or underscore.
    Replace illegal characters with `_`, and prepend a leading `_` if necessary.

    The string must contain at least one Unicode letter (matching the regex ``\\p{L}``),
    `_`, `!`, `?`, or number, but must not be `None`, `<NA>`, `N/A`, or `-`.

    Args:
        input_value: input value

    Raises:
        InputError: if the input cannot be transformed to an xsd:ID

    Returns:
        An xsd ID compatible string based on the input value

    Examples:
        ```python
        result = xmllib.make_xsd_compatible_id("0_Universität_Basel")
        # result == "_0_Universit_t_Basel"
        ```
    """
    if not is_nonempty_value_internal(input_value):
        raise_input_error(MessageInfo(f"The input '{input_value}' cannot be transformed to an xsd:ID"))
    # if the start of string is neither letter nor underscore, add an underscore
    res = regex.sub(r"^(?=[^A-Za-z_])", "_", str(input_value))
    # replace all illegal characters by underscore
    res = regex.sub(r"[^\w_\-.]", "_", res, flags=regex.ASCII)
    return res


def make_xsd_compatible_id_with_uuid(input_value: str | float | int) -> str:
    """
    An xsd:ID may not contain all types of special characters,
    and it must start with a letter or underscore.
    Replace illegal characters with `_`, and prepend a leading `_` if necessary.
    Additionally, add a UUID at the end.
    The UUID will be different each time the function is called.

    The string must contain at least one Unicode letter (matching the regex ``\\p{L}``),
    `_`, `!`, `?`, or number, but must not be `None`, `<NA>`, `N/A`, or `-`.

    Args:
        input_value: input value

    Raises:
        InputError: if the input cannot be transformed to an xsd:ID

    Returns:
        an xsd ID based on the input value, with a UUID attached.

    Examples:
        ```python
        result = xmllib.make_xsd_compatible_id_with_uuid("Universität_Basel")
        # result == "Universit_t_Basel_88f5cd0b-f333-4174-9030-65900b17773d"
        ```
    """
    res = make_xsd_compatible_id(input_value)
    _uuid = uuid.uuid4()
    res = f"{res}_{_uuid}"
    return res


def create_list_from_string(string: str, separator: str) -> list[str]:
    """
    Creates a list from a string.
    Trailing and leading whitespaces are removed from the list items.

    Args:
        string: input string
        separator: The character that separates the different values in the string.
            For example, a comma or newline.

    Returns:
        The list that results from splitting the input string.
            If the original string is empty or consists only of whitespace characters, the resulting list will be empty.

    Raises:
        InputError: If the input value is not a string.

    Examples:
        ```python
        result = xmllib.create_non_empty_list_from_string(" One/  Two\\n/", "/")
        # result == ["One", "Two"]
        ```

        ```python
        result = xmllib.create_list_from_string("   \\n    ", "\\n")
        # result == []
        ```
    """
    if not isinstance(string, str):
        raise_input_error(
            MessageInfo(f"The input for this function must be a string. Your input is a {type(string).__name__}.")
        )
    return [strpd for x in string.split(separator) if (strpd := x.strip())]


def create_non_empty_list_from_string(
    string: str, separator: str, resource_id: str | None = None, prop_name: str | None = None
) -> list[str]:
    """
    Creates a list from a string.
    Trailing and leading whitespaces are removed from the list items.

    If the resulting list is empty it will raise an `InputError`.

    Args:
        string: input string
        separator: The character that separates the different values in the string.
            For example, a comma or newline.
        resource_id: If the ID of the resource is provided, a better error message can be composed
        prop_name: If the name of the property is provided, a better error message can be composed

    Returns:
        The list that results from splitting the input string.

    Raises:
        InputError: If the resulting list is empty.

    Examples:
        ```python
        result = xmllib.create_non_empty_list_from_string("One\\nTwo   ", "\\n")
        # result == ["One", "Two"]
        ```

        ```python
        result = xmllib.create_non_empty_list_from_string("   \\n/    ", "/")
        # raises InputError
        ```
    """
    lst = create_list_from_string(string, separator)
    if len(lst) == 0:
        msg_info = MessageInfo(
            message="The input for this function must result in a non-empty list. Your input results in an empty list.",
            resource_id=resource_id,
            prop_name=prop_name,
        )
        raise_input_error(msg_info)
    return lst


def clean_whitespaces_from_string(string: str) -> str:
    """
    Remove redundant whitespaces (space, `\\n`, `\\t`, etc.) and replace them with a single space.

    If the resulting string is empty, a warning will be printed.

    Args:
        string: input string

    Returns:
        The cleaned string.

    Examples:
        ```python
        result = xmllib.clean_whitespaces_from_string("\\t Text\\nafter newline")
        # result == "Text after newline"
        ```

        ```python
        result = xmllib.clean_whitespaces_from_string("      \\n\\t ")
        # result == ""
        # warns that the string is now empty
        ```
    """
    cleaned = regex.sub(r"\s+", " ", string).strip()
    if len(cleaned) == 0:
        emit_xmllib_input_warning(
            MessageInfo(
                "The entered string is empty after all redundant whitespaces were removed. An empty string is returned."
            )
        )
    return cleaned


def find_license_in_string(string: str) -> License | None:  # noqa: PLR0911 (too many return statements)
    """
    Checks if a string contains a license, and returns it.
    Returns None if no license was found.
    The case (upper case/lower case) is ignored.

    Look out: Your string should contain no more than 1 license.
    If it contains more, there is no guarantee which one will be returned.

    See [recommended licenses](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/recommended/)
    for details.

    Args:
        string: string to check

    Returns:
        `License` object or `None`

    Examples:
        ```python
        result = xmllib.find_license_in_string("CC BY")
        # result == LicenseRecommended.CC.BY
        ```

        ```python
        result = xmllib.find_license_in_string("Creative Commons Developing Nations 2.0 Generic Deed")
        # result == None
        ```

    Currently supported license formats:
        - "AI" -> LicenseRecommended.DSP.AI_GENERATED
        - "KI" -> LicenseRecommended.DSP.AI_GENERATED
        - "IA" -> LicenseRecommended.DSP.AI_GENERATED
        - "public domain" -> LicenseRecommended.DSP.PUBLIC_DOMAIN
        - "gemeinfrei" -> LicenseRecommended.DSP.PUBLIC_DOMAIN
        - "frei von Urheberrechten" -> LicenseRecommended.DSP.PUBLIC_DOMAIN
        - "urheberrechtsbefreit" -> LicenseRecommended.DSP.PUBLIC_DOMAIN
        - "libre de droits" -> LicenseRecommended.DSP.PUBLIC_DOMAIN
        - "domaine public" -> LicenseRecommended.DSP.PUBLIC_DOMAIN
        - "unknown" -> LicenseRecommended.DSP.UNKNOWN
        - "unbekannt" -> LicenseRecommended.DSP.UNKNOWN
        - "inconnu" -> LicenseRecommended.DSP.UNKNOWN
        - "CC BY" -> LicenseRecommended.CC.BY
        - "Creative Commons BY 4.0" -> LicenseRecommended.CC.BY
        - "CC 0 1.0" -> LicenseOther.Public.CC_0_1_0
        - "CC PDM 1.0" -> LicenseOther.Public.CC_PDM_1_0
        - "BORIS Standard License" -> LicenseOther.Various.BORIS_STANDARD
        - "LICENCE OUVERTE 2.0" -> LicenseOther.Various.FRANCE_OUVERTE
    """
    if lic := _get_already_parsed_license(string):
        return lic

    sep = r"[-_\p{Zs}]+"  # Zs = unicode category for space separator characters

    if regex.search(rf"\b(Creative{sep}Commons|CC){sep}0({sep}1\.0)?\b", string, flags=regex.IGNORECASE):
        return LicenseOther.Public.CC_0_1_0

    if regex.search(rf"\b(Creative{sep}Commons|CC){sep}PDM({sep}1\.0)?\b", string, flags=regex.IGNORECASE):
        return LicenseOther.Public.CC_PDM_1_0

    if match := regex.search(
        rf"\b(CC|Creative{sep}Commons)({sep}(BY|NC|ND|SA))*({sep}[\d\.]+)?\b", string, flags=regex.IGNORECASE
    ):
        return _find_cc_license(match.group(0))

    if regex.search(r"\b(AI|IA|KI)\b", string, flags=regex.IGNORECASE):
        return LicenseRecommended.DSP.AI_GENERATED

    rgx_public_domain = (
        rf"\b(public{sep}domain|gemeinfrei|frei{sep}von{sep}Urheberrechten|urheberrechtsbefreit|"
        rf"libre{sep}de{sep}droits|domaine{sep}public)\b"
    )
    if regex.search(rgx_public_domain, string, flags=regex.IGNORECASE):
        return LicenseRecommended.DSP.PUBLIC_DOMAIN

    if regex.search(r"\b(unknown|unbekannt|inconnu)\b", string, flags=regex.IGNORECASE):
        return LicenseRecommended.DSP.UNKNOWN

    if regex.search(
        rf"\b(BORIS|Bern{sep}Open{sep}Repository{sep}and{sep}Information{sep}System){sep}Standard{sep}License\b",
        string,
        flags=regex.IGNORECASE,
    ):
        return LicenseOther.Various.BORIS_STANDARD

    if regex.search(
        rf"\b(France{sep})?Licence{sep}ouverte({sep}2\.0)?\b",
        string,
        flags=regex.IGNORECASE,
    ):
        return LicenseOther.Various.FRANCE_OUVERTE

    return None


def _find_cc_license(string: str) -> License | None:  # noqa: PLR0911 (too many return statements)
    string = string.lower()
    if "by" not in string:
        return None
    if any((string.count("by") > 1, string.count("nd") > 1, string.count("sa") > 1, string.count("nc") > 1)):
        return None
    has_nc = "nc" in string
    has_nd = "nd" in string
    has_sa = "sa" in string
    if not any((has_nc, has_nd, has_sa)):
        return LicenseRecommended.CC.BY
    if not has_nc and has_nd and not has_sa:
        return LicenseRecommended.CC.BY_ND
    if not has_nc and not has_nd and has_sa:
        return LicenseRecommended.CC.BY_SA
    if has_nc and not has_nd and not has_sa:
        return LicenseRecommended.CC.BY_NC
    if has_nc and has_nd and not has_sa:
        return LicenseRecommended.CC.BY_NC_ND
    if has_nc and not has_nd and has_sa:
        return LicenseRecommended.CC.BY_NC_SA
    return None


def _get_already_parsed_license(string: str) -> License | None:
    already_parsed_dict: dict[str, License] = {
        r"http://rdfh\.ch/licenses/cc-by-4\.0": LicenseRecommended.CC.BY,
        r"http://rdfh\.ch/licenses/cc-by-sa-4\.0": LicenseRecommended.CC.BY_SA,
        r"http://rdfh\.ch/licenses/cc-by-nc-4\.0": LicenseRecommended.CC.BY_NC,
        r"http://rdfh\.ch/licenses/cc-by-nc-sa-4\.0": LicenseRecommended.CC.BY_NC_SA,
        r"http://rdfh\.ch/licenses/cc-by-nd-4\.0": LicenseRecommended.CC.BY_ND,
        r"http://rdfh\.ch/licenses/cc-by-nc-nd-4\.0": LicenseRecommended.CC.BY_NC_ND,
        r"http://rdfh\.ch/licenses/ai-generated": LicenseRecommended.DSP.AI_GENERATED,
        r"http://rdfh\.ch/licenses/unknown": LicenseRecommended.DSP.UNKNOWN,
        r"http://rdfh\.ch/licenses/public-domain": LicenseRecommended.DSP.PUBLIC_DOMAIN,
        r"http://rdfh\.ch/licenses/cc-0-1.0": LicenseOther.Public.CC_0_1_0,
        r"http://rdfh\.ch/licenses/cc-pdm-1.0": LicenseOther.Public.CC_PDM_1_0,
        r"http://rdfh\.ch/licenses/boris": LicenseOther.Various.BORIS_STANDARD,
        r"http://rdfh\.ch/licenses/open-licence-2.0": LicenseOther.Various.FRANCE_OUVERTE,
    }
    for rgx, lic in already_parsed_dict.items():
        if regex.search(rgx, string):
            return lic
    return None



import contextlib
import shutil
from dataclasses import dataclass
from pathlib import Path

E2E_TESTDATA = Path("testdata/e2e").absolute()


@dataclass(frozen=True)
class ArtifactDirs:
    sipi_images: Path
    tmp_sipi: Path
    tmp_ingest: Path
    ingest_db: Path


def get_artifact_dirs(_uuid: str) -> ArtifactDirs:
    dirs = {
        "sipi_images": E2E_TESTDATA / "images" / _uuid,
        "tmp_sipi": E2E_TESTDATA / "tmp-dsp-sipi" / _uuid,
        "tmp_ingest": E2E_TESTDATA / "tmp-dsp-ingest" / _uuid,
        "ingest_db": E2E_TESTDATA / "ingest-db" / _uuid,
    }
    for _dir in dirs.values():
        _dir.mkdir(parents=True)
    return ArtifactDirs(**dirs)


def remove_artifact_dirs(artifact_dirs: ArtifactDirs) -> None:
    for _dir in [artifact_dirs.sipi_images, artifact_dirs.tmp_sipi, artifact_dirs.tmp_ingest, artifact_dirs.ingest_db]:
        with contextlib.suppress(PermissionError):
            shutil.rmtree(_dir)



from pathlib import Path

from loguru import logger

LOGGER_SAVEPATH = (Path.home() / ".dsp-tools" / "logging.log").absolute()
WARNINGS_SAVEPATH = Path("warnings.log")


def logger_config() -> None:
    """
    This function configures the log files.
    Currently, there are three sinks:
    - logger.log in the home directory for development purposes
    - warnings.log in the cwd only with level warning and higher for the user (no stack-trace)
    - print output on the terminal, formatted the same as the warnings.log
    """
    # If this is not removed, the default formatting is also printed out on the terminal
    logger.remove()

    text_format = "<level>{time:YYYY-MM-DD HH:mm:ss.SSS} | {level: <8} | {message}</level>"
    rotation_size = "100 MB"
    retention_number = 30

    logger.add(
        sink=LOGGER_SAVEPATH,
        format=text_format,
        backtrace=True,
        diagnose=True,
        rotation=rotation_size,
        retention=retention_number,
    )

    logger.add(
        sink=WARNINGS_SAVEPATH,
        level="WARNING",
        format=text_format,
        backtrace=False,
        diagnose=False,
        rotation=rotation_size,
        retention=10,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class OntoResource:
    name: str
    super: list[str]
    labels: LanguageDict
    comments: LanguageDict | None
    cardinalities: list[ResourceCardinality] | None

    def serialise(self) -> dict[str, Any]:
        res: dict[str, Any] = {
            "name": self.name,
            "super": self.super,
            "labels": self.labels.serialise(),
        }
        if self.comments:
            res["comments"] = self.comments.serialise()
        if self.cardinalities:
            res["cardinalities"] = [x.serialise() for x in self.cardinalities]
        return res


@dataclass
class ResourceCardinality:
    propname: str
    cardinality: str
    gui_order: int

    def serialise(self) -> dict[str, str | int]:
        return {"propname": self.propname, "cardinality": self.cardinality, "gui_order": self.gui_order}


@dataclass
class OntoProperty:
    name: str
    super: list[str]
    object: str
    subject: str | None
    labels: LanguageDict
    comments: LanguageDict | None
    gui_element: str
    gui_attributes: GuiAttributes | None

    def serialise(self) -> dict[str, Any]:
        prop_dict: dict[str, Any] = {"name": self.name, "super": self.super, "object": self.object}
        if self.subject:
            prop_dict["subject"] = self.subject
        prop_dict["labels"] = self.labels.serialise()
        if self.comments:
            prop_dict["comments"] = self.comments.serialise()
        prop_dict["gui_element"] = self.gui_element
        if self.gui_attributes:
            prop_dict["gui_attributes"] = self.gui_attributes.serialise()
        return prop_dict


@dataclass
class LanguageDict:
    lang_dict: dict[str, str]

    def serialise(self) -> dict[str, str]:
        return self.lang_dict


@dataclass
class GuiAttributes:
    gui_attributes: dict[str, int | str | float]

    def serialise(self) -> dict[str, int | str | float]:
        return self.gui_attributes



from loguru import logger
from rdflib import Graph

from dsp_tools.commands.validate_data.models.api_responses import EnabledLicenseIris


def construct_allowed_licenses_shape(license_iris: EnabledLicenseIris) -> Graph:
    """Create a constraint detailing the allowed licences."""
    formatted_iris = [f"<{x}>" for x in license_iris.enabled_licenses]
    license_str = " ".join(formatted_iris)
    if len(formatted_iris) == 0:
        msg_str = '"You are only allowed to reference enabled licenses. No licenses are enabled for this project."'
    else:
        msg_str = (
            '"Please only use enabled licenses in your data. Consult the project information for enabled licenses."'
        )
    logger.debug("Constructing allowed licesnses shapes.")
    ttl_str = """
    @prefix sh:         <http://www.w3.org/ns/shacl#> .
    @prefix knora-api:  <http://api.knora.org/ontology/knora-api/v2#> .
    @prefix api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#> .
    
    api-shapes:FileValue_ClassShape
      a sh:NodeShape ;
      sh:targetClass knora-api:FileValue ;
      sh:property [
          a           sh:PropertyShape ;
          sh:path     knora-api:hasLicense ;
          sh:in       ( %(license_str)s ) ;
          sh:message  %(msg)s ;
          sh:severity  sh:Violation
                  ] .
    """ % {"license_str": license_str, "msg": msg_str}  # noqa: UP031 Use format specifiers instead of percent format
    g = Graph()
    g.parse(data=ttl_str, format="turtle")
    return g



from collections.abc import Iterator
from pathlib import Path

import pytest

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create_all import create_project
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from test.e2e.setup_testcontainers.ports import ExternalContainerPorts
from test.e2e.setup_testcontainers.setup import get_containers


@pytest.fixture(scope="module")
def container_ports() -> Iterator[ExternalContainerPorts]:
    with get_containers() as metadata:
        yield metadata.ports


@pytest.fixture(scope="module")
def creds(container_ports: ExternalContainerPorts) -> ServerCredentials:
    return ServerCredentials(
        "root@example.com",
        "test",
        f"http://0.0.0.0:{container_ports.api}",
        f"http://0.0.0.0:{container_ports.ingest}",
    )


@pytest.fixture(scope="module")
def api_url(container_ports: ExternalContainerPorts) -> str:
    return f"http://0.0.0.0:{container_ports.api}"


@pytest.fixture(scope="module")
def shacl_validator(api_url: str) -> ShaclValidator:
    return ShaclValidator(api_url)


@pytest.fixture(scope="module")
def create_generic_project(creds: ServerCredentials) -> None:
    assert create_project(Path("testdata/validate-data/generic/project.json"), creds)



from typing import Union

import numpy as np
import pandas as pd
import pytest

from dsp_tools.commands.excel2xml.propertyelement import PropertyElement
from dsp_tools.utils.data_formats import shared


def test_check_notna_good() -> None:
    na_values = [
        None,
        pd.NA,
        pd.NA,
        "",
        "  ",
        "-",
        ",",
        ".",
        "*",
        " ⳰",
        " ῀ ",  # noqa: RUF001 (ambiguous-unicode-character-string)
        " ῾ ",  # noqa: RUF001 (ambiguous-unicode-character-string)
        " \n\t ",
        "N/A",
        "n/a",
        "<NA>",
        "None",
        ["a", "b"],
        pd.array(["a", "b"]),
        np.array([0, 1]),
    ]
    for na_value in na_values:
        assert not shared.check_notna(na_value), f"Failed na_value: {na_value}"


def test_check_notna_problem() -> None:
    notna_values_orig: list[Union[str, int, float, bool]] = [
        1,
        0.1,
        True,
        False,
        "True",
        "False",
        r" \n\t ",
        "0",
        "_",
        "Ὅμηρος",
        "!",
        "?",
    ]
    notna_values_as_propelem = [PropertyElement(x) for x in notna_values_orig]
    for notna_value in notna_values_orig + notna_values_as_propelem:
        assert shared.check_notna(notna_value), f"Failed notna_value: {notna_value}"


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from lxml import etree


@pytest.fixture
def root_no_resources() -> etree._Element:
    return etree.fromstring("""
    <knora shortcode="0000"
       default-ontology="default">
    </knora>
    """)


@pytest.fixture
def root_with_2_resources() -> etree._Element:
    return etree.fromstring("""
    <knora shortcode="0000"
       default-ontology="default">
        <resource label="The only resource"
                  restype=":minimalResource"
                  id="the_only_resource">
        </resource>
        
        <link label="a link object" id="link_obj">
            <resptr-prop name="hasLinkTo">
                <resptr permissions="open">the_only_resource</resptr>
            </resptr-prop>
        </link>
    </knora>
    """)


@pytest.fixture
def resource_no_values() -> etree._Element:
    return etree.fromstring('<resource label="lbl" restype=":Class" id="resource_no_values"/>')


@pytest.fixture
def resource_empty_with_permission() -> etree._Element:
    return etree.fromstring("""
    <resource 
        label="lbl" 
        restype=":Class" 
        id="resource_empty_permissions"
        permissions="open"
    >
    </resource>
    """)


@pytest.fixture
def resource_with_file_value() -> etree._Element:
    return etree.fromstring("""
    <resource label="lbl" restype=":Class" id="resource_with_file_value">
        <bitstream>
            testdata/bitstreams/test.wav    
        </bitstream>
    </resource>
    """)


@pytest.fixture
def resource_with_iiif() -> etree._Element:
    return etree.fromstring("""
    <resource label="lbl" restype=":Class" id="resource_with_iiif">
        <iiif-uri>
            https://iiif.uri/full.jpg
        </iiif-uri>
    </resource>
    """)


@pytest.fixture
def resource_with_migration_data() -> etree._Element:
    return etree.fromstring("""
    <resource 
        label="lbl" 
        restype=":Class" 
        id="resource_with_migration_data"
        ark="ark"
        iri="iri"
        creation_date="2019-01-09T15:45:54.502951Z"
    >
    </resource>
    """)


@pytest.fixture
def resource_with_value() -> etree._Element:
    return etree.fromstring("""
        <resource label="lbl" restype=":Class" id="resource_with_value">
            <boolean-prop name=":hasProp">
                <boolean>true</boolean>
            </boolean-prop>
        </resource>""")


@pytest.fixture
def resource_region() -> etree._Element:
    return etree.fromstring("""
    <region label="lbl" id="resource_region">
        <color-prop name="hasColor">
            <color>#5d1f1e</color>
        </color-prop>
        <resptr-prop name="isRegionOf">
            <resptr>target_region_of</resptr>
        </resptr-prop>
        <geometry-prop name="hasGeometry">
            <geometry>
                {}
            </geometry>
        </geometry-prop>
        <text-prop name="hasComment">
            <text encoding="xml"> Comment</text>
        </text-prop>
    </region>
    """)


@pytest.fixture
def resource_link() -> etree._Element:
    return etree.fromstring("""
    <link label="lbl" id="resource_link">
        <resptr-prop name="hasLinkTo">
            <resptr permissions="open"> target</resptr>
        </resptr-prop>
    </link>
    """)


@pytest.fixture
def resource_video_segment() -> etree._Element:
    return etree.fromstring("""
    <video-segment label="lbl" id="resource_video_segment">
        <isSegmentOf>target</isSegmentOf>
        <hasSegmentBounds segment_start="0.1" segment_end="0.234"/>
        <hasTitle>Title</hasTitle>
        <hasComment>Comment</hasComment>
        <hasDescription> Description</hasDescription>
        <hasKeyword>Keyword 1</hasKeyword>
        <hasKeyword>Keyword 2</hasKeyword>
        <relatesTo>relates_to_id</relatesTo>
    </video-segment>
    """)


@pytest.fixture
def resource_audio_segment() -> etree._Element:
    return etree.fromstring("""
    <audio-segment label="lbl" id="resource_audio_segment">
        <isSegmentOf>target</isSegmentOf>
        <hasSegmentBounds permissions="open" segment_start="0.1" segment_end="0.234"/>
        <hasTitle comment="Cmt"> Title</hasTitle>
        <hasComment><p>Comment</p></hasComment>
        <hasDescription> <p>Description 1</p></hasDescription>
        <hasDescription>Description 2</hasDescription>
        <hasKeyword>Keyword </hasKeyword>
        <relatesTo>relates_to_id</relatesTo>
    </audio-segment>
    """)



import pytest

from dsp_tools.utils.data_formats.iri_util import is_resource_iri


def test_is_resource_iri() -> None:
    assert is_resource_iri("https://rdfh.ch/0000/PKzNC8MFQT6NAwR3OaQEKw")
    assert is_resource_iri("http://rdfh.ch/0000/PKzNC8MFQT6NAwR3OaQEKw")
    assert is_resource_iri("https://rdfh.ch/0000/4HATaCuPQeir-L8LyhVj-Q")
    assert is_resource_iri("https://rdfh.ch/0000/31eSNdrwQkae9_0c7-753w")


def test_is_not_resource_iri() -> None:
    assert not is_resource_iri("https://knora.org/0000/PKzNC8MFQT6NAwR3OaQEKw")
    assert not is_resource_iri("https://rdfh.ch/0000/PKzNC8MFQT6NAwR3OaQ")
    assert not is_resource_iri("www.rdfh.ch/0000/PKzNC8MFQT6NAwR3OaQEKw")


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import importlib.resources
from copy import deepcopy
from pathlib import Path

import pandas as pd
import regex
from loguru import logger
from lxml import etree

from dsp_tools.error.exceptions import InputError
from dsp_tools.error.xsd_validation_error_msg import XSDValidationMessage
from dsp_tools.error.xsd_validation_error_msg import get_xsd_validation_message_str
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_RED
from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT

separator = "\n    "
list_separator = "\n    - "


def parse_and_clean_xml_file(input_file: Path) -> etree._Element:
    root = _parse_xml_file(input_file)
    root = _remove_comments_from_element_tree(root)
    _validate_xml_with_schema(root)
    print("The XML file is syntactically correct.")
    return _transform_into_localnames(root)


def parse_and_validate_xml_file(input_file: Path | str) -> bool:
    root = _parse_xml_file(input_file)
    data_xml = _remove_comments_from_element_tree(root)
    return _validate_xml_with_schema(data_xml)


def _parse_xml_file(input_file: str | Path) -> etree._Element:
    parser = etree.XMLParser(remove_comments=True, remove_pis=True)
    try:
        return etree.parse(source=input_file, parser=parser).getroot()
    except etree.XMLSyntaxError as err:
        logger.error(f"The XML file contains the following syntax error: {err.msg}")
        raise InputError(f"The XML file contains the following syntax error: {err.msg}") from None


def _transform_into_localnames(root: etree._Element) -> etree._Element:
    """Removes the namespace of the tags."""
    tree = deepcopy(root)
    for elem in tree.iter():
        elem.tag = etree.QName(elem).localname
    return tree


def _remove_comments_from_element_tree(input_tree: etree._Element) -> etree._Element:
    """Removes comments and processing instructions."""
    root = deepcopy(input_tree)
    for c in root.xpath("//comment()"):
        c.getparent().remove(c)
    for c in root.xpath("//processing-instruction()"):
        c.getparent().remove(c)
    return root


def _validate_xml_with_schema(xml: etree._Element) -> bool:
    """Requires a cleaned (no comments) XML, but with the namespaces."""
    if errors := _validate_xml_tree_against_schema(xml):
        problem_msg = _get_validation_error_print_out(errors)
        logger.error(problem_msg)
        raise InputError(problem_msg)
    return True


def _validate_xml_tree_against_schema(data_xml: etree._Element) -> etree.XMLSchema | None:
    schema_res = importlib.resources.files("dsp_tools").joinpath("resources/schema/data.xsd")
    with schema_res.open(encoding="utf-8") as schema_file:
        xmlschema = etree.XMLSchema(etree.parse(schema_file))
    if not xmlschema.validate(data_xml):
        return xmlschema
    return None


def _get_validation_error_print_out(xmlschema: etree.XMLSchema) -> str:
    error_msg = "The XML file cannot be uploaded due to the following validation error(s):"
    for error in xmlschema.error_log:
        error_msg += f"{separator}Line {error.line}: {_beautify_err_msg(error.message)}"
    return error_msg


def _beautify_err_msg(err_msg: str) -> str:
    err_msg = err_msg.replace("{https://dasch.swiss/schema}", "")
    new_msg_for_duplicate_res_id = (
        "The resource ID '\\1' is not valid. IDs must be unique across the entire file. "
        "The function make_xsd_compatible_id() assists you in creating IDs."
    )
    rgx_for_duplicate_res_id = (
        r"Element '(?:resource|link|video-segment|audio-segment|region)', attribute 'id': '(.+?)' "
        r"is not a valid value of the atomic type 'xs:ID'."
    )
    err_msg = regex.sub(rgx_for_duplicate_res_id, new_msg_for_duplicate_res_id, err_msg)
    return err_msg


def validate_root_emit_user_message(root: etree._Element, save_path: Path) -> None:
    validation_errors = _validate_root_get_validation_messages(root)
    if validation_errors:
        _emit_validation_errors(validation_errors, save_path)


def _validate_root_get_validation_messages(data_xml: etree._Element) -> list[XSDValidationMessage] | None:
    if errors := _validate_xml_tree_against_schema(data_xml):
        return _reformat_validation_errors(errors.error_log)
    return None


def _emit_validation_errors(validation_errors: list[XSDValidationMessage], save_path: Path) -> None:
    header_msg = f"During the XSD Schema validation the following {len(validation_errors)} error(s) were found: "
    print(BACKGROUND_BOLD_RED, header_msg, RESET_TO_DEFAULT)
    logger.error(header_msg)
    if len(validation_errors) > 50:
        save_path = save_path / "xsd_validation_errors.csv"
        message_dicts = [vars(x) for x in validation_errors]
        df = pd.DataFrame.from_records(message_dicts)
        df.to_csv(save_path, index=False)
        msg = f"Due to the large number of errors they are saved in the file '{save_path}'."
        print(BOLD_RED + msg, RESET_TO_DEFAULT)
        logger.error(msg)
    else:
        for one_msg in validation_errors:
            msg_str = get_xsd_validation_message_str(one_msg)
            print(BOLD_RED, msg_str, RESET_TO_DEFAULT)
            logger.error(msg_str)


def _reformat_validation_errors(log: etree._ListErrorLog) -> list[XSDValidationMessage]:
    return [_reformat_error_message_str(err.message, err.line) for err in log]


def _reformat_error_message_str(msg: str, line_number: int) -> XSDValidationMessage:
    element, attrib = None, None
    msg = msg.replace("{https://dasch.swiss/schema}", "")
    first, message = msg.split(":", maxsplit=1)
    if ele_found := regex.search(r"Element '(.*?)'", first):
        element = ele_found.group(1)
    if attrib_found := regex.search(r"attribute '(.*?)'", first):
        attrib = attrib_found.group(1)
    if " is not a valid value of the atomic type 'xs:ID'." in message:
        if found := regex.search(r"'.*?'", message):
            id_ = found.group(0)
        else:
            id_ = ""
        message = f"The provided resource id {id_} is either not a valid xsd:ID or not unique in the file."
    else:
        message = regex.sub(r"\[facet .+\] ", "", message)
        message = regex.sub(r"pattern ('.+')", "pattern for this value", message).strip()
    return XSDValidationMessage(line_number=line_number, element=element, attribute=attrib, message=message)



# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest
from rdflib import Graph

from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import OneNode
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.utils import reformat_onto_iri
from dsp_tools.commands.validate_data.validate_data import _check_for_unknown_resource_classes
from dsp_tools.commands.validate_data.validate_data import _get_all_onto_classes
from dsp_tools.commands.validate_data.validate_data import _get_msg_str_unknown_classes_in_data
from dsp_tools.commands.validate_data.validate_data import _get_validation_status
from dsp_tools.commands.validate_data.validate_data import _make_list_lookup
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from test.unittests.commands.validate_data.constants import PREFIXES

ONTO_STR = "http://0.0.0.0:3333/ontology/9999/onto/v2#"
NON_EXISTING_ONTO = "http://0.0.0.0:3333/ontology/9999/non-existent/v2#"

CLASSES_IN_ONTO = {
    f"{ONTO_STR}One",
    f"{KNORA_API_STR}LinkObj",
    f"{KNORA_API_STR}Region",
    f"{KNORA_API_STR}AudioSegment",
    f"{KNORA_API_STR}VideoSegment",
}

PREFIXED_IN_ONTO = {reformat_onto_iri(x) for x in CLASSES_IN_ONTO}


def _get_rdf_graphs(data_graph: Graph) -> RDFGraphs:
    ttl = f"""{PREFIXES}
    onto:One a owl:Class ;
            knora-api:isResourceClass true ;
            knora-api:canBeInstantiated true .

    knora-api:TextValue a owl:Class ;
            knora-api:isValueClass true .
    """
    onto_g = Graph()
    onto_g.parse(data=ttl, format="turtle")
    knora_subset = Graph()
    knora_subset.parse("testdata/validate-data/knora-api-subset.ttl")
    return RDFGraphs(
        data=data_graph, ontos=onto_g, cardinality_shapes=Graph(), content_shapes=Graph(), knora_api=knora_subset
    )


class TestFindUnknownClasses:
    def test_check_for_unknown_resource_classes_data_ok(self):
        ttl = f"""{PREFIXES}
        <http://data/identical_text_different_prop> a onto:One ;
                onto:testSimpleText <http://data/textValue> .

        <http://data/textValue> a knora-api:TextValue ;
                knora-api:valueAsString "Text"^^xsd:string .
        """
        g = Graph()
        g.parse(data=ttl, format="turtle")
        graphs = _get_rdf_graphs(g)
        result = _check_for_unknown_resource_classes(graphs, CLASSES_IN_ONTO)
        assert not result

    def test_check_for_unknown_resource_classes_data_wrong(self):
        ttl = f"""{PREFIXES}
        <http://data/identical_text_different_prop> a onto:NonExistent ;
                onto:testSimpleText <http://data/textValue> .

        <http://data/textValue> a knora-api:TextValue ;
                knora-api:valueAsString "Text"^^xsd:string .
        """
        g = Graph()
        g.parse(data=ttl, format="turtle")
        graphs = _get_rdf_graphs(g)
        used_iris = {f"{ONTO_STR}NonExistent"}
        result = _check_for_unknown_resource_classes(graphs, used_iris)
        assert isinstance(result, UnknownClassesInData)
        assert result.unknown_classes == {"onto:NonExistent"}
        assert result.defined_classes == PREFIXED_IN_ONTO
        expected_msg = (
            "Your data uses resource classes that do not exist in the ontologies in the database.\n"
            "The following classes that are used in the data are unknown: onto:NonExistent\n"
            "The following classes exist in the uploaded ontologies: "
            "AudioSegment, LinkObj, Region, VideoSegment, onto:One"
        )
        res_msg = _get_msg_str_unknown_classes_in_data(result)
        assert res_msg == expected_msg

    def test_check_for_unknown_resource_classes_data_prefix_non_existent(self):
        ttl = f"""{PREFIXES}
        @prefix  non-existent: <{NON_EXISTING_ONTO}> .

        <http://data/identical_text_different_prop> a non-existent:One ;
                onto:testSimpleText <http://data/textValue> .

        <http://data/textValue> a knora-api:TextValue ;
                knora-api:valueAsString "Text"^^xsd:string .
        """
        g = Graph()
        g.parse(data=ttl, format="turtle")
        graphs = _get_rdf_graphs(g)
        used_iris = {f"{NON_EXISTING_ONTO}One"}
        result = _check_for_unknown_resource_classes(graphs, used_iris)
        assert isinstance(result, UnknownClassesInData)
        assert result.unknown_classes == {"non-existent:One"}
        assert result.defined_classes == PREFIXED_IN_ONTO
        expected_msg = (
            "Your data uses ontologies that don't exist in the database.\n"
            "The following ontologies that are used in the data are unknown: non-existent\n"
            "The following ontologies are uploaded: onto"
        )
        res_msg = _get_msg_str_unknown_classes_in_data(result)
        assert res_msg == expected_msg

    def test_get_all_onto_classes(self):
        graphs = _get_rdf_graphs(Graph())
        all_user_facing_classes = _get_all_onto_classes(graphs)
        assert all_user_facing_classes == CLASSES_IN_ONTO


def test_make_list_lookup():
    all_lists = [
        OneList("IRI1", "list1", [OneNode("l1n1", "IRIl1n1"), OneNode("l1n2", "IRIl1n2")]),
        OneList("IRI2", "list2", [OneNode("l2n1", "IRIl2n1")]),
    ]
    result = _make_list_lookup(all_lists)
    expected_lookup = {
        ("", "IRIl1n1"): "IRIl1n1",
        ("", "IRIl1n2"): "IRIl1n2",
        ("", "IRIl2n1"): "IRIl2n1",
        ("list1", "l1n1"): "IRIl1n1",
        ("list1", "l1n2"): "IRIl1n2",
        ("list2", "l2n1"): "IRIl2n1",
    }
    assert result.lists == expected_lookup


class TestGetValidationStatus:
    def test_not_prod_problems(self):
        problems = SortedProblems([InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [], [], [])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert not validation_passed

    def test_not_prod_problems_and_unexpected(self):
        problems = SortedProblems([InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [], [], ["unex"])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert not validation_passed

    def test_not_prod_warnings(self):
        problems = SortedProblems([], [InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [], [])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert validation_passed

    def test_not_prod_info(self):
        problems = SortedProblems([], [], [InputProblem(ProblemType.GENERIC, "", "", "", Severity.VIOLATION)], [])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert validation_passed

    def test_not_prod_unexpected(self):
        problems = SortedProblems([], [], [], ["unexpected"])
        validation_passed = _get_validation_status(problems, is_on_prod=False)
        assert not validation_passed


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest

from dsp_tools.xmllib.internal.input_converters import numeric_entities


@pytest.mark.parametrize(
    ("original", "expected"),
    [
        ("a &nbsp; a", "a &#160; a"),
        ("a &#160; a", "a &#160; a"),
        ("a &#x22; a", "a &#x22; a"),
        ("a &quot; &amp; &apos; &lt; &gt; a", "a &quot; &amp; &apos; &lt; &gt; a"),
        ("aäö&;", "aäö&;"),
    ],
)
def test_numeric_entities(original: str, expected: str) -> None:
    assert numeric_entities(original) == expected
